# -*- coding: utf-8 -*-
# Copyright 2021, CS GROUP - France, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import re

from eodag.api.product.drivers.base import AssetPatterns, DatasetDriver

logger = logging.getLogger("eodag.driver.generic")


class GenericDriver(DatasetDriver):
    """Generic default Driver"""

    #: list of patterns to match asset keys and roles
    ASSET_KEYS_PATTERNS_ROLES: list[AssetPatterns] = [
        # data
        {
            "pattern": re.compile(
                (
                    r"^(?:.*[/\\])?([^/\\]+)"
                    r"(\.jp2|\.tiff?|\.dat|\.nc|\.grib2?|"
                    r"\.zarr|\.nat|\.covjson|\.parquet|\.zip|\.tar|\.gz)$"
                ),
                re.IGNORECASE,
            ),
            "roles": ["data"],
        },
        # metadata
        {
            "pattern": re.compile(
                r"^(?:.*[/\\])?([^/\\]+)(\.xml|\.xsd|\.safe|\.json)$", re.IGNORECASE
            ),
            "roles": ["metadata"],
        },
        # thumbnail
        {
            "pattern": re.compile(
                r"^(?:.*[/\\])?(thumbnail)(\.jpg|\.jpeg|\.png)$", re.IGNORECASE
            ),
            "roles": ["thumbnail"],
        },
        # quicklook
        {
            "pattern": re.compile(
                r"^(?:.*[/\\])?([^/\\]+-ql|preview|quick-?look)(\.jpg|\.jpeg|\.png)$",
                re.IGNORECASE,
            ),
            "roles": ["overview"],
        },
        # default
        {"pattern": re.compile(r"^(?:.*[/\\])?([^/\\]+)$"), "roles": ["auxiliary"]},
    ]



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Optional

from eodag import EOProduct
from eodag.plugins.download.aws import AwsDownload
from eodag.utils import _deprecated


@_deprecated(
    reason="Plugin that was used in creodias_s3 provider configuration, but not anymore",
    version="3.10.0",
)
class CreodiasS3Download(AwsDownload):
    """
    Download on creodias s3 from their VMs (extension of :class:`~eodag.plugins.download.aws.AwsDownload`)

    :param provider: provider name
    :param config: Download plugin configuration:

        * :attr:`~eodag.config.PluginConfig.type` (``str``) (**mandatory**): CreodiasS3Download
        * :attr:`~eodag.config.PluginConfig.base_uri` (``str``) (**mandatory**): s3 endpoint url
        * :attr:`~eodag.config.PluginConfig.s3_bucket` (``str``) (**mandatory**): bucket where the products can be found
        * :attr:`~eodag.config.PluginConfig.ssl_verify` (``bool``): if the ssl certificates should be
          verified in requests; default: ``True``
    """

    def _get_bucket_names_and_prefixes(
        self,
        product: EOProduct,
        asset_filter: Optional[str],
        ignore_assets: bool,
        complementary_url_keys: list[str],
    ) -> list[tuple[str, Optional[str]]]:
        """
        Retrieves the bucket names and path prefixes for the assets

        :param product: product for which the assets shall be downloaded
        :param asset_filter: text for which the assets should be filtered
        :param ignore_assets: if product instead of individual assets should be used
        :return: tuples of bucket names and prefixes
        """
        # if assets are defined, use them instead of scanning product.location
        if len(product.assets) > 0 and not ignore_assets:
            bucket_names_and_prefixes = super()._get_bucket_names_and_prefixes(
                product, asset_filter, ignore_assets, complementary_url_keys
            )
        else:
            # if no assets are given, use productIdentifier to get S3 path for download
            s3_url = "s3:/" + product.properties["productIdentifier"]
            bucket_names_and_prefixes = [
                self.get_product_bucket_name_and_prefix(product, s3_url)
            ]
        return bucket_names_and_prefixes



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
import unittest
from contextlib import contextmanager
from datetime import datetime
from importlib.resources import files as res_files
from tempfile import TemporaryDirectory

import click
from click.testing import CliRunner
from faker import Faker
from packaging import version

from eodag.api.search_result import SearchResult
from eodag.utils import GENERIC_PRODUCT_TYPE
from tests import TEST_RESOURCES_PATH
from tests.context import (
    DEFAULT_ITEMS_PER_PAGE,
    AuthenticationError,
    MisconfiguredError,
    NoMatchingProductType,
    download,
    eodag,
    search_crunch,
)
from tests.units import test_core
from tests.utils import mock, no_blanks, write_eodag_conf_with_fake_credentials


class TestEodagCli(unittest.TestCase):
    @contextmanager
    def user_conf(self, conf_file="user_conf.yml", content=b"key: to unused conf"):
        """Utility method"""
        with self.runner.isolated_filesystem():
            with open(conf_file, "wb") as fd:
                fd.write(
                    content if isinstance(content, bytes) else content.encode("utf-8")
                )
            yield conf_file

    def setUp(self):
        super(TestEodagCli, self).setUp()
        self.runner = CliRunner()
        self.faker = Faker()

        # Mock home and eodag conf directory to tmp dir
        self.tmp_home_dir = TemporaryDirectory()
        self.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, return_value=self.tmp_home_dir.name
        )
        self.expanduser_mock.start()

        # create eodag conf dir in tmp home dir
        eodag_conf_dir = os.path.join(self.tmp_home_dir.name, ".config", "eodag")
        os.makedirs(eodag_conf_dir, exist_ok=False)
        # use empty config file with fake credentials in order to have full
        # list for tests and prevent providers to be pruned
        write_eodag_conf_with_fake_credentials(
            os.path.join(eodag_conf_dir, "eodag.yml")
        )

    def tearDown(self):
        super(TestEodagCli, self).tearDown()
        # stop Mock and remove tmp config dir
        self.expanduser_mock.stop()
        self.tmp_home_dir.cleanup()
        # reset logging
        logger = logging.getLogger("eodag")
        logger.handlers = []
        logger.level = 0

    def test_eodag_without_args(self):
        """Calling eodag without arguments should print help message"""
        result = self.runner.invoke(eodag)
        self.assertIn(
            "Usage: eodag [OPTIONS] COMMAND1 [ARGS]... [COMMAND2 [ARGS]...]...",
            result.output,
        )
        # Exit status 2 with no_args_is_help starting click >= 8.2.0
        self.assertIn(result.exit_code, (0, 2))

    def test_eodag_with_only_verbose_opt(self):
        """Calling eodag only with -v option should print error message"""
        result = self.runner.invoke(eodag, ["-v"])
        self.assertIn("Error: Missing command.", result.output)
        self.assertNotEqual(result.exit_code, 0)

    def test_eodag_cli_version(self):
        """Calling eodag version should return the version"""
        result = self.runner.invoke(eodag, ["version"])
        version_str = re.search(r"eodag .* version (.+)\s*$", result.output).group(1)
        self.assertGreater(version.parse(version_str), version.parse("2.0.0"))

    def test_eodag_search_without_args(self):
        """Calling eodag search subcommand without arguments should print help message and return error code"""  # noqa
        result = self.runner.invoke(eodag, ["search"])
        with click.Context(search_crunch) as ctx:
            self.assertEqual(
                no_blanks(result.output),
                no_blanks(
                    "".join(
                        (
                            "Give me some work to do. See below for how to do that:",
                            ctx.get_help(),
                        )
                    )
                ),
            )
        self.assertNotEqual(result.exit_code, 0)

    def test_eodag_search_without_producttype_arg(self):
        """Calling eodag search without -p | --productType should print the help message and return error code"""  # noqa
        start_date = self.faker.date()
        end_date = self.faker.date()
        result = self.runner.invoke(eodag, ["search", "-s", start_date, "-e", end_date])
        with click.Context(search_crunch) as ctx:
            self.assertEqual(
                no_blanks(result.output),
                no_blanks(
                    "".join(
                        (
                            "Give me some work to do. See below for how to do that:",
                            ctx.get_help(),
                        )
                    )
                ),
            )
        self.assertNotEqual(result.exit_code, 0)

    def test_eodag_search_with_conf_file_inexistent(self):
        """Calling eodag search with --conf | -f set to a non-existent file should print error message"""  # noqa
        conf_file = "does_not_exist.yml"
        result = self.runner.invoke(
            eodag, ["search", "--conf", conf_file, "-p", "whatever"]
        )
        expect_output = "Error: Invalid value for '-f' / '--conf': Path '{}' does not exist.".format(  # noqa
            conf_file
        )
        self.assertTrue(
            expect_output in result.output or expect_output.replace("'", '"')
        )
        self.assertNotEqual(result.exit_code, 0)

    def test_eodag_search_with_max_cloud_out_of_range(self):
        """Calling eodag search with -c | --maxCloud set to a value < 0 or > 100 should print error message"""  # noqa
        with self.user_conf() as conf_file:
            for max_cloud in (110, -1):
                result = self.runner.invoke(
                    eodag,
                    ["search", "--conf", conf_file, "-p", "whatever", "-c", max_cloud],
                )
                expect_output = (
                    "Error: Invalid value for '-c' / '--cloudCover': {} is not in the"
                    " valid range of 0 to 100."
                ).format(max_cloud)
                self.assertTrue(
                    expect_output in result.output or expect_output.replace("'", '"')
                )
                self.assertNotEqual(result.exit_code, 0)

    def test_eodag_search_bbox_invalid(self):
        """Calling eodag search with -b | --bbox set with less than 4 params should print error message"""  # noqa
        with self.user_conf() as conf_file:
            result = self.runner.invoke(
                eodag, ["search", "--conf", conf_file, "-p", "whatever", "-b", 1, 2]
            )
            self.assertIn("-b", result.output)
            self.assertIn("requires 4 arguments", result.output)
            self.assertNotEqual(result.exit_code, 0)

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_bbox_valid(self, dag):
        """Calling eodag search with --bbox argument valid"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            self.runner.invoke(
                eodag,
                ["search", "--conf", conf_file, "-p", product_type, "-b", 1, 43, 2, 44],
            )
            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                page=1,
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                cloudCover=None,
                geometry={"lonmin": 1, "latmin": 43, "lonmax": 2, "latmax": 44},
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                productType=product_type,
                id=None,
                locations=None,
                count=False,
            )

    def test_eodag_search_geom_wkt_invalid(self):
        """Calling eodag search with -g | --geom set with invalit WKT geometry string"""  # noqa
        with self.user_conf() as conf_file:
            result = self.runner.invoke(
                eodag,
                ["search", "--conf", conf_file, "-p", "whatever", "-g", "not a wkt"],
            )
            # GEOSException for shapely >= 2.0
            assert "WKTReadingError" in str(result) or "GEOSException" in str(result)
            self.assertNotEqual(result.exit_code, 0)

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_geom_wkt_valid(self, dag):
        """Calling eodag search with --geom WKT argument valid"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    product_type,
                    "-g",
                    "POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                ],
            )
            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                items_per_page=DEFAULT_ITEMS_PER_PAGE,
                page=1,
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                cloudCover=None,
                geometry="POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                productType=product_type,
                id=None,
                locations=None,
                count=False,
            )

    def test_eodag_search_bbox_geom_mutually_exclusive(self):
        """Calling eodag search with both --geom and --box"""  # noqa
        with self.user_conf() as conf_file:
            result = self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    "whatever",
                    "-b",
                    1,
                    2,
                    3,
                    4,
                    "-g",
                    "a wkt",
                ],
            )
            self.assertIn("Illegal usage", result.output)
            self.assertNotEqual(result.exit_code, 0)

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_storage_arg(self, dag):
        """Calling eodag search with specified result filename without .geojson extension"""  # noqa
        with self.user_conf() as conf_file:
            api_obj = dag.return_value
            api_obj.search.return_value = SearchResult([mock.MagicMock() * 2], 2)
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    "whatever",
                    "--storage",
                    "results",
                ],
            )
            api_obj.serialize.assert_called_with(
                api_obj.search.return_value, filename="results.geojson"
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_with_cruncher(self, dag):
        """Calling eodag search with --cruncher arg should call crunch method of search result"""  # noqa
        with self.user_conf() as conf_file:
            api_obj = dag.return_value
            api_obj.search.return_value = SearchResult([mock.MagicMock() * 2], 2)

            product_type = "whatever"
            cruncher = "FilterLatestIntersect"
            criteria = dict(
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                geometry=None,
                cloudCover=None,
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                productType=product_type,
                id=None,
                locations=None,
            )
            self.runner.invoke(
                eodag,
                ["search", "-f", conf_file, "-p", product_type, "--cruncher", cruncher],
            )

            search_results = api_obj.search.return_value
            crunch_results = api_obj.crunch.return_value

            # Assertions
            dag.assert_called_once_with(
                user_conf_file_path=conf_file, locations_conf_path=None
            )
            api_obj.search.assert_called_once_with(
                count=False, items_per_page=DEFAULT_ITEMS_PER_PAGE, page=1, **criteria
            )
            api_obj.crunch.assert_called_once_with(
                search_results, search_criteria=criteria, **{cruncher: {}}
            )
            api_obj.serialize.assert_called_with(
                crunch_results, filename="search_results.geojson"
            )

            # Call with a cruncher taking arguments
            cruncher = "FilterOverlap"
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "-f",
                    conf_file,
                    "-p",
                    product_type,
                    "--cruncher",
                    cruncher,
                    "--cruncher-args",
                    cruncher,
                    "minimum_overlap",
                    "10",
                ],
            )
            api_obj.crunch.assert_called_with(
                search_results,
                search_criteria=criteria,
                **{cruncher: {"minimum_overlap": "10"}},
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_and_download(self, dag):
        """Calling eodag search with --download argument should directly download search results"""
        with self.user_conf() as conf_file:
            api_obj = dag.return_value
            search_results = api_obj.search.return_value = SearchResult(
                [mock.MagicMock() * 2], 2
            )

            product_type = "whatever"
            criteria = dict(
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                geometry=None,
                cloudCover=None,
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                productType=product_type,
                id=None,
                locations=None,
            )
            self.runner.invoke(
                eodag,
                ["search", "-f", conf_file, "-p", product_type, "download"],
            )

            # Assertions
            self.assertEqual(dag.call_count, 2)
            dag.assert_any_call(user_conf_file_path=conf_file, locations_conf_path=None)
            api_obj.search.assert_called_once_with(
                count=False, items_per_page=DEFAULT_ITEMS_PER_PAGE, page=1, **criteria
            )
            api_obj.download_all.assert_called_once_with(
                search_results, output_dir=None
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_all(self, dag):
        """Calling eodag search with --bbox argument valid"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    product_type,
                    "-g",
                    "POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                    "--all",
                ],
            )
            api_obj = dag.return_value
            api_obj.search_all.assert_called_once_with(
                items_per_page=None,
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                cloudCover=None,
                geometry="POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                productType=product_type,
                id=None,
                locations=None,
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_query(self, dag):
        """Calling eodag search with --query argument"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    product_type,
                    "--query",
                    "foo=1&bar=2&bar=3",
                ],
            )
            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                page=1,
                items_per_page=20,
                geometry=None,
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                cloudCover=None,
                productType=product_type,
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                id=None,
                foo="1",
                bar=["2", "3"],
                locations=None,
                count=False,
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_locations(self, dag):
        """Calling eodag search with --locations argument"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    product_type,
                    "--locations",
                    "country=FRA&continent=Africa",
                ],
            )
            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                page=1,
                items_per_page=20,
                geometry=None,
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=None,
                cloudCover=None,
                productType=product_type,
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                id=None,
                locations={"country": "FRA", "continent": "Africa"},
                count=False,
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_start_date(self, dag):
        """Calling eodag search with --start argument"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            start_date_str = "2022-01-01"
            start_date_datetime = datetime.strptime(
                start_date_str, "%Y-%m-%d"
            ).isoformat()
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    product_type,
                    "--start",
                    start_date_str,
                ],
            )
            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                page=1,
                items_per_page=20,
                geometry=None,
                startTimeFromAscendingNode=start_date_datetime,
                completionTimeFromAscendingNode=None,
                cloudCover=None,
                productType=product_type,
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                id=None,
                locations=None,
                count=False,
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_stop_date(self, dag):
        """Calling eodag search with --end argument"""
        with self.user_conf() as conf_file:
            product_type = "whatever"
            stop_date_str = "2022-01-01"
            stop_date_datetime = datetime.strptime(
                stop_date_str, "%Y-%m-%d"
            ).isoformat()
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    product_type,
                    "--end",
                    stop_date_str,
                    "--count",
                ],
            )
            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                page=1,
                items_per_page=20,
                geometry=None,
                startTimeFromAscendingNode=None,
                completionTimeFromAscendingNode=stop_date_datetime,
                cloudCover=None,
                productType=product_type,
                instrument=None,
                platform=None,
                platformSerialIdentifier=None,
                processingLevel=None,
                sensorType=None,
                id=None,
                locations=None,
                count=True,
            )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_search_locs(self, dag):
        """Calling eodag search with --locs argument"""
        with self.user_conf() as conf_file:
            locs_file = os.path.join(TEST_RESOURCES_PATH, "file_locations_override.yml")

            product_type = "whatever"
            self.runner.invoke(
                eodag,
                [
                    "search",
                    "--conf",
                    conf_file,
                    "--locs",
                    locs_file,
                    "-p",
                    product_type,
                ],
            )

            dag.assert_called_once_with(
                user_conf_file_path=conf_file,
                locations_conf_path=locs_file,
            )

    def test_eodag_list_product_type_ok(self):
        """Calling eodag list without provider should return all supported product types"""
        all_supported_product_types = [
            pt
            for pt, provs in test_core.TestCore.SUPPORTED_PRODUCT_TYPES.items()
            if len(provs) != 0 and pt != GENERIC_PRODUCT_TYPE
        ]
        result = self.runner.invoke(eodag, ["list", "--no-fetch"])
        self.assertEqual(result.exit_code, 0)
        for pt in all_supported_product_types:
            self.assertIn(pt, result.output)

    def test_eodag_list_product_type_with_provider_ok(self):
        """Calling eodag list with provider should return all supported product types of specified provider"""  # noqa
        for provider in test_core.TestCore.SUPPORTED_PROVIDERS:
            provider_supported_product_types = [
                pt
                for pt, provs in test_core.TestCore.SUPPORTED_PRODUCT_TYPES.items()
                if provider in provs
                if pt != GENERIC_PRODUCT_TYPE
            ]
            result = self.runner.invoke(eodag, ["list", "-p", provider, "--no-fetch"])
            self.assertEqual(result.exit_code, 0)
            for pt in provider_supported_product_types:
                self.assertIn(
                    pt,
                    result.output,
                    f"{pt} was not found in {provider} supported product types",
                )

    def test_eodag_list_product_type_with_provider_ko(self):
        """Calling eodag list with unsupported provider should fail and print a list of available providers"""  # noqa
        provider = "random"
        result = self.runner.invoke(eodag, ["list", "-p", provider, "--no-fetch"])
        self.assertEqual(result.exit_code, 1)
        self.assertIn("Unsupported provider. You may have a typo", result.output)
        self.assertIn(
            f"Available providers: {', '.join(test_core.TestCore.SUPPORTED_PROVIDERS)}",
            result.output,
        )

    @mock.patch("eodag.cli.EODataAccessGateway.fetch_product_types_list", autospec=True)
    def test_eodag_list_product_type_fetch(self, mock_fetch_product_types_list):
        """Calling eodag list should fetch for new product types depending on passed option"""
        result = self.runner.invoke(eodag, ["list", "--no-fetch"])
        self.assertEqual(result.exit_code, 0)
        assert not mock_fetch_product_types_list.called

        result = self.runner.invoke(eodag, ["list"])
        self.assertEqual(result.exit_code, 0)
        mock_fetch_product_types_list.assert_called_once_with(mock.ANY, provider=None)

        result = self.runner.invoke(eodag, ["list", "-p", "peps"])
        self.assertEqual(result.exit_code, 0)
        mock_fetch_product_types_list.assert_called_with(mock.ANY, provider="peps")
        self.assertEqual(mock_fetch_product_types_list.call_count, 2)

    @mock.patch("eodag.cli.EODataAccessGateway.list_product_types", autospec=True)
    @mock.patch("eodag.cli.EODataAccessGateway.guess_product_type", autospec=True)
    def test_eodag_guess_product_type_ok(
        self, mock_guess_product_type, mock_list_product_types
    ):
        """Calling eodag list with one or several valid product type feature(s) should return
        all supported product types with this (these) feature(s) among the ones of its provider
        and with or without fetching provider according to the command.
        """
        provider = "peps"
        mock_guess_product_type.return_value = ["foo", "bar"]
        mock_list_product_types.return_value = [
            {"ID": "foo", "title": "this is foo"},
            {"ID": "bar", "title": "this is bar"},
            {"ID": "baz", "title": "this is baz"},
        ]

        result = self.runner.invoke(
            eodag,
            ["list", "-p", provider, "--platformSerialIdentifier", "S2A", "--no-fetch"],
        )
        self.assertEqual(result.exit_code, 0)

        self.assertIn("foo", result.output)
        self.assertIn("bar", result.output)
        self.assertNotIn("baz", result.output)

        mock_list_product_types.assert_called_with(
            mock.ANY, provider=provider, fetch_providers=False
        )

    @mock.patch(
        "eodag.cli.EODataAccessGateway.guess_product_type",
        autospec=True,
        side_effect=NoMatchingProductType(),
    )
    def test_eodag_guess_product_type_ko(self, mock_guess_product_type):
        """Calling eodag list with invalid product type feature(s) should print a
        'no matching' type message and return error code.
        """
        result = self.runner.invoke(
            eodag,
            ["list", "--platformSerialIdentifier", "fake_identifier", "--no-fetch"],
        )
        self.assertIn(
            "No product type match the following criteria you provided:\n",
            result.output,
        )
        self.assertNotEqual(result.exit_code, 0)

    @mock.patch(
        "eodag.cli.EODataAccessGateway.discover_product_types",
        autospec=True,
        return_value={},
    )
    def test_eodag_discover_product_types(self, mock_discover_product_types):
        """Calling eodag discover should fetch providers for new product types"""
        origin_dir = os.getcwd()
        try:
            os.chdir(self.tmp_home_dir.name)
            default_output_path = os.path.join(
                self.tmp_home_dir.name, "ext_product_types.json"
            )
            self.assertFalse(os.path.isfile(default_output_path))

            # call without any arg
            result = self.runner.invoke(eodag, ["discover"])
            self.assertEqual(result.exit_code, 0)
            mock_discover_product_types.assert_called_once_with(mock.ANY)
            self.assertTrue(os.path.isfile(default_output_path))

            # call with provider
            result = self.runner.invoke(eodag, ["discover", "-p", "peps"])
            self.assertEqual(result.exit_code, 0)
            mock_discover_product_types.assert_called_with(mock.ANY, provider="peps")
            self.assertEqual(mock_discover_product_types.call_count, 2)
            os.remove(default_output_path)

            other_file_path = os.path.join(self.tmp_home_dir.name, "toto")
            # call with filename without extension
            self.assertFalse(os.path.isfile(other_file_path))
            self.assertFalse(os.path.isfile(f"{other_file_path}.json"))
            result = self.runner.invoke(eodag, ["discover", "--storage", "toto"])
            self.assertEqual(result.exit_code, 0)
            mock_discover_product_types.assert_called_with(mock.ANY)
            self.assertEqual(mock_discover_product_types.call_count, 3)
            self.assertFalse(os.path.isfile(other_file_path))
            self.assertTrue(os.path.isfile(f"{other_file_path}.json"))
            os.remove(f"{other_file_path}.json")

            # call with filename with extension
            self.assertFalse(os.path.isfile(f"{other_file_path}.json"))
            result = self.runner.invoke(eodag, ["discover", "--storage", "toto.json"])
            self.assertEqual(result.exit_code, 0)
            mock_discover_product_types.assert_called_with(mock.ANY)
            self.assertEqual(mock_discover_product_types.call_count, 4)
            self.assertTrue(os.path.isfile(f"{other_file_path}.json"))
            os.remove(f"{other_file_path}.json")
        finally:
            os.chdir(origin_dir)

    def test_eodag_download_no_search_results_arg(self):
        """Calling eodag download without a path to a search result should fail"""
        result = self.runner.invoke(eodag, ["download"])
        with click.Context(download) as ctx:
            self.assertEqual(
                no_blanks(result.output),
                no_blanks(
                    "".join(
                        (
                            "Nothing to do (no search results file or stac item provided)",
                            ctx.get_help(),
                        )
                    )
                ),
            )
        self.assertEqual(result.exit_code, 1)

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_download_ok(self, dag):
        """Calling eodag download with all args well formed succeed"""
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        config_path = os.path.join(TEST_RESOURCES_PATH, "file_config_override.yml")
        dag.return_value.download_all.return_value = ["/fake_path"]
        result = self.runner.invoke(
            eodag,
            ["download", "--search-results", search_results_path, "-f", config_path],
        )
        dag.assert_called_once_with(user_conf_file_path=config_path)
        dag.return_value.deserialize_and_register.assert_called_once_with(
            search_results_path
        )
        self.assertEqual(dag.return_value.download_all.call_count, 1)
        self.assertEqual("Downloaded /fake_path\n", result.output)

        # Testing the case when no downloaded path is returned
        dag.return_value.download_all.return_value = [None]
        result = self.runner.invoke(
            eodag,
            ["download", "--search-results", search_results_path, "-f", config_path],
        )
        self.assertEqual(
            "A file may have been downloaded but we cannot locate it\n", result.output
        )

        # Testing output directory
        output_dir = os.path.join(self.tmp_home_dir.name, "downloads")
        result = self.runner.invoke(
            eodag,
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--output-dir",
                output_dir,
            ],
        )
        dag.return_value.download_all.assert_called_with(
            mock.ANY, output_dir=output_dir
        )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_download_ko(self, dag):
        """Calling eodag download with all args well formed fails"""
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        config_path = os.path.join(TEST_RESOURCES_PATH, "file_config_override.yml")
        dag.return_value.download_all.return_value = []
        result = self.runner.invoke(
            eodag,
            ["download", "--search-results", search_results_path, "-f", config_path],
        )
        self.assertEqual(
            "Error during download, a file may have been downloaded but we cannot locate it\n",
            result.output,
        )

    @mock.patch("eodag.cli.EODataAccessGateway", autospec=True)
    def test_eodag_download_stac_items(self, dag):
        """Calling eodag download with --stac-item argument"""
        fake_result = SearchResult([mock.MagicMock() * 2], 2)
        dag.return_value.import_stac_items.return_value = fake_result
        with self.user_conf() as conf_file:
            self.runner.invoke(
                eodag,
                [
                    "download",
                    "--conf",
                    conf_file,
                    "--stac-item",
                    "foo",
                    "--stac-item",
                    "bar",
                ],
            )
            dag.return_value.import_stac_items.assert_called_once_with(
                ["foo", "bar"],
            )
            dag.return_value.download_all.assert_called_once_with(
                fake_result, output_dir=None
            )

    def test_eodag_download_missingcredentials(self):
        """Calling eodag download with missing credentials must raise MisconfiguredError"""
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        default_conf_file = str(
            res_files("eodag") / "resources" / "user_conf_template.yml"
        )
        result = self.runner.invoke(
            eodag,
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                default_conf_file,
            ],
        )
        self.assertEqual(result.exit_code, 1)
        self.assertIsInstance(result.exception, MisconfiguredError)

    @mock.patch("eodag.plugins.download.http.HTTPDownload.download", autospec=True)
    def test_eodag_download_wrongcredentials(self, download):
        """Calling eodag download with wrong credentials must raise AuthenticationError"""
        # This is not an end-to-end test so we have to manually raise the error down
        # to HTTPDownload.download. This is indeed the download plugin of PEPS which
        # is used here since the GeoJSON results were obtained from this provider.
        download.side_effect = AuthenticationError
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        result = self.runner.invoke(
            eodag,
            ["download", "--search-results", search_results_path],
            # We override the default (empty) credentials with dummy values not
            # to raise a MisconfiguredError.
            env={
                "EODAG__PEPS__AUTH__CREDENTIALS__USERNAME": "dummy",
                "EODAG__PEPS__AUTH__CREDENTIALS__PASSWORD": "dummy",
            },
        )
        self.assertEqual(result.exit_code, 1)
        self.assertIsInstance(result.exception, AuthenticationError)
        self.assertEqual(download.call_count, 1)

    @mock.patch("eodag.api.product._product.EOProduct.get_quicklook", autospec=True)
    def test_eodag_download_quicklooks(self, mock_get_quicklook):
        """Calling eodag download with --quicklooks argument"""
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result_peps.geojson"
        )
        config_path = os.path.join(TEST_RESOURCES_PATH, "file_config_override.yml")
        mock_get_quicklook.return_value = "/fake_path"
        result = self.runner.invoke(
            eodag,
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--quicklooks",
            ],
        )
        self.assertIn(
            "Flag 'quicklooks' specified, downloading only quicklooks of products\n",
            result.output,
        )

        # 2 quicklooks are called in search_results_path
        self.assertEqual(mock_get_quicklook.call_count, 2)
        self.assertIn("Downloaded /fake_path\n", result.output)

        # change output dir
        output_dir = os.path.join(self.tmp_home_dir.name, "quicklooks")
        result = self.runner.invoke(
            eodag,
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--quicklooks",
                "--output-dir",
                output_dir,
            ],
        )
        mock_get_quicklook.assert_called_with(mock.ANY, output_dir=output_dir)

        # Testing the case when no quicklook path is returned
        mock_get_quicklook.return_value = None
        result = self.runner.invoke(
            eodag,
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--quicklooks",
            ],
        )
        self.assertIn(
            "A quicklook may have been downloaded but we cannot locate it. "
            "Increase verbosity for more details: `eodag -v download [OPTIONS]`",
            result.output,
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import operator
from typing import TYPE_CHECKING, Any

from eodag.plugins.crunch.base import Crunch

logger = logging.getLogger("eodag.crunch.property")

if TYPE_CHECKING:
    from eodag.api.product import EOProduct


class FilterProperty(Crunch):
    """FilterProperty cruncher

    Filter products, retaining only those whose property match criteria

    :param config: Crunch configuration, must contain :

        * ``<property>`` ``(Any)`` (**mandatory**): property key from ``product.properties``, associated to its filter
          value
        * ``operator`` (``str``): Operator used for filtering (one of ``lt,le,eq,ne,ge,gt``). Default is ``eq``
    """

    def proceed(
        self, products: list[EOProduct], **search_params: Any
    ) -> list[EOProduct]:
        """Execute crunch: Filter products, retaining only those that match property filtering

        :param products: A list of products resulting from a search
        :returns: The filtered products
        """
        operator_name = self.config.__dict__.pop("operator", "eq") or "eq"
        try:
            operator_method = getattr(operator, operator_name)
        except AttributeError:
            logger.warning(
                "Unknown operator `%s`, should be one of `lt,le,eq,ne,ge,gt`",
                operator_name,
            )
            return products

        if len(self.config.__dict__.keys()) != 1:
            logger.warning("One property is needed for filtering, filtering disabled.")
            return products

        property_key = next(iter(self.config.__dict__))
        property_value = self.config.__dict__.get(property_key)

        logger.debug(
            "Start filtering for products matching operator.%s(product.properties['%s'], %s)",
            operator_name,
            property_key,
            property_value,
        )
        filtered: list[EOProduct] = []
        add_to_filtered = filtered.append

        for product in products:
            if property_key not in product.properties.keys():
                logger.warning(
                    f"{property_key} not found in {product}.properties, product skipped",
                )
                continue
            if operator_method(product.properties[property_key], property_value):
                add_to_filtered(product)

        logger.info("Finished filtering products. %s resulting products", len(filtered))
        return filtered



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qs, urlparse

import requests
from requests.auth import AuthBase
from requests.exceptions import RequestException

from eodag.plugins.authentication import Authentication
from eodag.utils import HTTP_REQ_TIMEOUT, USER_AGENT
from eodag.utils.exceptions import AuthenticationError, TimeOutError

if TYPE_CHECKING:
    from requests import PreparedRequest


class HttpQueryStringAuth(Authentication):
    """An Authentication plugin using HTTP query string parameters.

    This plugin sends credentials as query-string parameters.

    :param provider: provider name
    :param config: Authentication plugin configuration:

        * :attr:`~eodag.config.PluginConfig.type` (``str``) (**mandatory**): HttpQueryStringAuth
        * :attr:`~eodag.config.PluginConfig.auth_uri` (``str``): used to check the credentials
          given in the configuration

    Using :class:`~eodag.plugins.download.http.HTTPDownload` a download link
    ``http://example.com?foo=bar`` will become
    ``http://example.com?foo=bar&apikey=XXX&otherkey=YYY`` if associated to the following
    configuration::

        provider:
            credentials:
                apikey: XXX
                otherkey: YYY

    The plugin is configured as follows in the providers config file::

        provider:
            ...
            auth:
                plugin: HttpQueryStringAuth
                auth_uri: 'http://example.com?foo=bar'
                ...
            ...

    If ``auth_uri`` is specified (optional), it will be used to check credentials through
    :meth:`~eodag.plugins.authentication.query_string.HttpQueryStringAuth.authenticate`
    """

    def authenticate(self) -> AuthBase:
        """Authenticate"""
        self.validate_config_credentials()

        auth = QueryStringAuth(**self.config.credentials)

        auth_uri = getattr(self.config, "auth_uri", None)
        ssl_verify = getattr(self.config, "ssl_verify", True)

        if auth_uri:
            try:
                response = requests.get(
                    auth_uri,
                    timeout=HTTP_REQ_TIMEOUT,
                    headers=USER_AGENT,
                    auth=auth,
                    verify=ssl_verify,
                )
                response.raise_for_status()
            except requests.exceptions.Timeout as exc:
                raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
            except RequestException as e:
                raise AuthenticationError("Could no authenticate", str(e)) from e

        return auth


class QueryStringAuth(AuthBase):
    """ "QueryStringAuth custom authentication class to be used with requests module"""

    def __init__(self, **parse_args: Any) -> None:
        self.parse_args = parse_args

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        """Perform the actual authentication"""
        parts = urlparse(str(request.url))
        query_dict = parse_qs(parts.query)
        query_dict.update(self.parse_args)
        url_without_args = parts._replace(query="").geturl()

        request.prepare_url(url_without_args, query_dict)
        return request



# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.cs-soprasteria.com/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import logging
from typing import Any, Callable, TypeVar

logger = logging.getLogger("eodag.cache")

R = TypeVar("R")


def instance_cached_method(
    maxsize: int = 128,
) -> Callable[[Callable[..., R]], Callable[..., R]]:
    """
    Decorator to cache an instance method using functools.lru_cache on a per-instance basis.

    This decorator creates a separate LRU cache for each instance of the class,
    ensuring that cached results are not shared across instances.

    The cache stores up to `maxsize` entries (default 128), which matches the
    default cache size of functools.lru_cache. This default provides a good balance
    between memory consumption and cache hit rate for most use cases.

    :param maxsize: Maximum number of cached calls to store per instance.
                    Defaults to 128, consistent with functools.lru_cache.
    :return: Decorated method with per-instance caching enabled.
    """

    def decorator(method: Callable[..., R]) -> Callable[..., R]:
        @functools.wraps(method)
        def wrapper(self: Any, *args: Any, **kwargs: Any) -> R:
            cache_name = f"_cached_{method.__name__}"
            if not hasattr(self, cache_name):
                cached = functools.lru_cache(maxsize=maxsize)(
                    method.__get__(self, type(self))
                )
                setattr(self, cache_name, cached)

            cached_func = getattr(self, cache_name)
            before_hits = cached_func.cache_info().hits
            result = cached_func(*args, **kwargs)
            after_hits = cached_func.cache_info().hits

            if after_hits > before_hits:
                logger.debug(
                    f"Cache hit for {method.__qualname__} with args={args}, kwargs={kwargs}"
                )
            return result

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ast
import os
import unittest
from datetime import datetime, timedelta, timezone
from tempfile import TemporaryDirectory
from unittest import mock

import geojson
import responses
from dateutil.parser import isoparse
from ecmwfapi.api import ANONYMOUS_APIKEY_VALUES
from shapely.geometry import shape

from eodag.utils import deepcopy
from tests.context import (
    DEFAULT_DOWNLOAD_WAIT,
    DEFAULT_MISSION_START_DATE,
    ONLINE_STATUS,
    USER_AGENT,
    USGS_TMPFILE,
    AuthenticationError,
    EODataAccessGateway,
    EOProduct,
    NotAvailableError,
    PluginManager,
    PreparedSearch,
    SearchResult,
    USGSAuthExpiredError,
    USGSError,
    get_geometry_from_various,
    load_default_config,
    path_to_uri,
    urlsplit,
)


class BaseApisPluginTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(BaseApisPluginTest, cls).setUpClass()
        providers_config = load_default_config()
        cls.plugins_manager = PluginManager(providers_config)
        # Mock home and eodag conf directory to tmp dir
        cls.tmp_home_dir = TemporaryDirectory()
        expanduser_mock_side_effect = (
            lambda *x: x[0]
            .replace("~user", cls.tmp_home_dir.name)
            .replace("~", cls.tmp_home_dir.name)
        )
        cls.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, side_effect=expanduser_mock_side_effect
        )
        cls.expanduser_mock.start()

    @classmethod
    def tearDownClass(cls):
        super(BaseApisPluginTest, cls).tearDownClass()
        # stop Mock and remove tmp config dir
        cls.expanduser_mock.stop()
        cls.tmp_home_dir.cleanup()

    def get_search_plugin(self, product_type=None, provider=None):
        return next(
            self.plugins_manager.get_search_plugins(
                product_type=product_type, provider=provider
            )
        )


class TestApisPluginEcmwfApi(BaseApisPluginTest):
    def setUp(self):
        self.provider = "ecmwf"
        self.api_plugin = self.get_search_plugin(provider=self.provider)
        self.query_dates = {
            "startTimeFromAscendingNode": "2022-01-01T00:00:00.000Z",
            "completionTimeFromAscendingNode": "2022-01-02T00:00:00.000Z",
        }
        self.product_type = "TIGGE_CF_SFC"
        self.product_type_params = {
            "ecmwf:class": "ti",
            "ecmwf:dataset": "tigge",
            "ecmwf:expver": "prod",
            "ecmwf:type": "cf",
            "ecmwf:levtype": "sfc",
            "ecmwf:origin": "ecmwf",
            "ecmwf:grid": "0.5/0.5",
            "ecmwf:param": "59/134/136/146/147/151/165/166/167/168/172/176/177/179/189/235/"
            + "228002/228039/228139/228141/228144/228164/228228",
            "ecmwf:step": 0,
            "ecmwf:time": "00:00",
            "ecmwf:target": "output",
        }
        self.custom_query_params = {
            "ecmwf:origin": "ecmwf",
            "ecmwf:levtype": "sfc",
            "ecmwf:number": "1",
            "ecmwf:expver": "prod",
            "ecmwf:dataset": "tigge",
            "ecmwf:step": "0",
            "ecmwf:grid": "2/2",
            "ecmwf:param": "228164",  # total cloud cover parameter
            "ecmwf:time": "00",
            "ecmwf:type": "cf",
            "ecmwf:class": "ti",
        }

    def test_plugins_apis_ecmwf_query_dates_missing(self):
        """Ecmwf.query must use default dates if missing"""
        # given start & stop
        results, _ = self.api_plugin.query(
            productType=self.product_type,
            startTimeFromAscendingNode="2020-01-01",
            completionTimeFromAscendingNode="2020-01-02",
        )
        eoproduct = results[0]
        self.assertEqual(
            eoproduct.properties["startTimeFromAscendingNode"],
            "2020-01-01T00:00:00.000Z",
        )
        self.assertEqual(
            eoproduct.properties["completionTimeFromAscendingNode"],
            "2020-01-02T00:00:00.000Z",
        )

        # missing start & stop
        results, _ = self.api_plugin.query(
            productType=self.product_type,
        )
        eoproduct = results[0]
        self.assertIn(
            eoproduct.properties["startTimeFromAscendingNode"],
            DEFAULT_MISSION_START_DATE,
        )
        # less than 10 seconds should have passed since the product was created
        self.assertLess(
            datetime.now(timezone.utc)
            - isoparse(eoproduct.properties["completionTimeFromAscendingNode"]),
            timedelta(seconds=10),
            "stop date should have been created from datetime.now",
        )

        # missing start & stop and plugin.product_type_config set (set in core._prepare_search)
        self.api_plugin.config.product_type_config = {
            "productType": self.product_type,
            "missionStartDate": "1985-10-26",
            "missionEndDate": "2015-10-21",
        }
        results, _ = self.api_plugin.query(
            productType=self.product_type,
        )
        eoproduct = results[0]
        self.assertEqual(
            eoproduct.properties["startTimeFromAscendingNode"],
            "1985-10-26T00:00:00.000Z",
        )
        self.assertEqual(
            eoproduct.properties["completionTimeFromAscendingNode"],
            "2015-10-21T00:00:00.000Z",
        )

    def test_plugins_apis_ecmwf_query_without_producttype(self):
        """
        EcmwfApi.query must build a EOProduct from input parameters without product type.
        For test only, result cannot be downloaded.
        """
        results, count = self.api_plugin.query(**self.query_dates)
        assert count == 1
        eoproduct = results[0]
        assert eoproduct.geometry.bounds == (-180.0, -90.0, 180.0, 90.0)
        assert (
            eoproduct.properties["startTimeFromAscendingNode"]
            == self.query_dates["startTimeFromAscendingNode"]
        )
        assert (
            eoproduct.properties["completionTimeFromAscendingNode"]
            == self.query_dates["completionTimeFromAscendingNode"]
        )
        assert eoproduct.properties["title"] == eoproduct.properties["id"]
        assert eoproduct.properties["title"].startswith("MARS___")
        assert eoproduct.location.startswith("http")

    def test_plugins_apis_ecmwf_query_with_producttype(self):
        """EcmwfApi.query must build a EOProduct from input parameters with predefined product type"""
        results, _ = self.api_plugin.query(
            **self.query_dates, productType=self.product_type, geometry=[1, 2, 3, 4]
        )
        eoproduct = results[0]
        assert eoproduct.properties["title"].startswith(self.product_type)
        assert eoproduct.geometry.bounds == (1.0, 2.0, 3.0, 4.0)
        # check if product_type_params is a subset of eoproduct.properties
        assert self.product_type_params.items() <= eoproduct.properties.items()
        params = deepcopy(self.query_dates)
        params["productType"] = self.product_type
        params["ecmwf:param"] = "tcc"

        # product type default settings can be overwritten using search kwargs
        results, _ = self.api_plugin.query(**params)
        eoproduct = results[0]
        assert eoproduct.properties["ecmwf:param"] == "tcc"

    def test_plugins_apis_ecmwf_query_with_custom_producttype(self):
        """EcmwfApi.query must build a EOProduct from input parameters with custom product type"""
        results, _ = self.api_plugin.query(
            **self.query_dates,
            **self.custom_query_params,
        )
        eoproduct = results[0]
        assert eoproduct.properties["title"].startswith(
            "%s_%s_%s"
            % (
                self.custom_query_params["ecmwf:dataset"].upper(),
                self.custom_query_params["ecmwf:type"].upper(),
                self.custom_query_params["ecmwf:levtype"].upper(),
            )
        )
        # check if custom_query_params is a subset of eoproduct.properties
        for param in self.custom_query_params:
            try:
                # for numeric values
                assert eoproduct.properties[param] == ast.literal_eval(
                    self.custom_query_params[param]
                )
            except Exception:
                assert eoproduct.properties[param] == self.custom_query_params[param]

    @mock.patch("ecmwfapi.api.Connection.call", autospec=True)
    def test_plugins_apis_ecmwf_authenticate(self, mock_connection_call):
        """EcmwfApi.authenticate must return a credentials dict"""
        # auth using ~/.ecmwfapirc
        auth_dict = self.api_plugin.authenticate()
        assert (
            auth_dict["key"],
            auth_dict["url"],
            auth_dict["email"],
        ) == ANONYMOUS_APIKEY_VALUES

        # auth using eodag credentials
        credentials = {
            "username": "foo",
            "password": "bar",
            "url": "http://foo.bar.baz",
        }
        self.api_plugin.config.credentials = credentials
        auth_dict = self.api_plugin.authenticate()
        assert auth_dict["email"] == credentials["username"]
        assert auth_dict["key"] == credentials["password"]
        assert auth_dict["url"] == self.api_plugin.config.auth_endpoint
        del self.api_plugin.config.credentials

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.sessions.Session.request",
        autospec=True,
    )
    @mock.patch(
        "eodag.api.core.EODataAccessGateway.fetch_product_types_list", autospec=True
    )
    @mock.patch("ecmwfapi.api.ECMWFService.execute", autospec=True)
    @mock.patch("ecmwfapi.api.ECMWFDataServer.retrieve", autospec=True)
    @mock.patch("ecmwfapi.api.Connection.call", autospec=True)
    def test_plugins_apis_ecmwf_download(
        self,
        mock_connection_call,
        mock_ecmwfdataserver_retrieve,
        mock_ecmwfservice_execute,
        mock_fetch_product_types_list,
        mock_auth_session_request,
    ):
        """EcmwfApi.download must call the appriate ecmwf api service"""

        dag = EODataAccessGateway()
        dag.set_preferred_provider("ecmwf")
        output_data_path = os.path.join(os.path.expanduser("~"), "data")

        # public dataset request
        def create_empty_file_for_public_dataset(*args, **kwargs):
            with open(args[1]["target"], "x"):
                pass

        mock_ecmwfdataserver_retrieve.side_effect = create_empty_file_for_public_dataset
        results = dag.search(
            **self.query_dates,
            **self.custom_query_params,
        )
        eoproduct = results[0]
        expected_path = os.path.join(
            output_data_path, "%s" % eoproduct.properties["title"]
        )
        arg_path = os.path.join(
            output_data_path,
            "%s" % eoproduct.properties["title"],
            "%s.grib" % eoproduct.properties["title"],
        )
        path = eoproduct.download(output_dir=output_data_path)
        mock_ecmwfservice_execute.assert_not_called()
        mock_ecmwfdataserver_retrieve.assert_called_once_with(
            mock.ANY,  # ECMWFDataServer instance
            dict(
                target=arg_path,
                **geojson.loads(urlsplit(eoproduct.remote_location).query),
            ),
        )
        assert path == expected_path
        assert path_to_uri(expected_path) == eoproduct.location

        mock_ecmwfservice_execute.reset_mock()
        mock_ecmwfdataserver_retrieve.reset_mock()

        # operation archive request
        def create_empty_file_for_operation_archive(*args, **kwargs):
            with open(args[2], "x"):
                pass

        mock_ecmwfservice_execute.side_effect = create_empty_file_for_operation_archive
        operation_archive_custom_query_params = self.custom_query_params.copy()
        operation_archive_custom_query_params.pop("ecmwf:dataset")
        results = dag.search(
            **self.query_dates,
            **operation_archive_custom_query_params,
        )
        eoproduct = results[0]
        expected_path = os.path.join(
            output_data_path, "%s" % eoproduct.properties["title"]
        )
        arg_path = os.path.join(
            output_data_path,
            "%s" % eoproduct.properties["title"],
            "%s.grib" % eoproduct.properties["title"],
        )
        path = eoproduct.download(output_dir=output_data_path)
        download_request = geojson.loads(urlsplit(eoproduct.remote_location).query)
        download_request.pop("dataset", None)
        mock_ecmwfservice_execute.assert_called_once_with(
            mock.ANY,  # ECMWFService instance
            dict(
                **download_request,
            ),
            arg_path,
        )
        mock_ecmwfdataserver_retrieve.assert_not_called()
        assert path == expected_path
        assert path_to_uri(expected_path) == eoproduct.location

        mock_ecmwfservice_execute.reset_mock()
        mock_ecmwfdataserver_retrieve.reset_mock()

        # download again
        eoproduct.download(output_dir=output_data_path)
        mock_ecmwfservice_execute.assert_not_called()
        mock_ecmwfdataserver_retrieve.assert_not_called()

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.sessions.Session.request",
        autospec=True,
    )
    @mock.patch(
        "eodag.api.core.EODataAccessGateway.fetch_product_types_list", autospec=True
    )
    @mock.patch("ecmwfapi.api.ECMWFDataServer.retrieve", autospec=True)
    @mock.patch("ecmwfapi.api.Connection.call", autospec=True)
    def test_plugins_apis_ecmwf_download_all(
        self,
        mock_connection_call,
        mock_ecmwfdataserver_retrieve,
        mock_fetch_product_types_list,
        mock_auth_session_request,
    ):
        """EcmwfApi.download_all must call the appropriate ecmwf api service"""

        dag = EODataAccessGateway()
        dag.set_preferred_provider("ecmwf")

        eoproducts = SearchResult([])

        # public dataset request
        params = deepcopy(self.query_dates)
        params.update(self.custom_query_params)
        params["ecmwf:accuracy"] = "bar"
        results = dag.search(**params)
        eoproducts.extend(results)
        params["ecmwf:accuracy"] = "baz"
        results = dag.search(**params)
        eoproducts.extend(results)
        assert len(eoproducts) == 2

        paths = dag.download_all(
            eoproducts, output_dir=os.path.join(os.path.expanduser("~"), "data")
        )
        assert mock_ecmwfdataserver_retrieve.call_count == 2
        assert len(paths) == 2


class TestApisPluginUsgsApi(BaseApisPluginTest):
    SCENE_SEARCH_RETURN = {
        "data": {
            "results": [
                {
                    "browse": [
                        {
                            "browsePath": "https://path/to/quicklook.jpg",
                            "thumbnailPath": "https://path/to/thumbnail.jpg",
                        },
                        {},
                        {},
                    ],
                    "cloudCover": "77.46",
                    "entityId": "LC81780382020041LGN00",
                    "displayId": "LC08_L1GT_178038_20200210_20200224_01_T2",
                    "spatialBounds": {
                        "type": "Polygon",
                        "coordinates": [
                            [
                                [28.03905, 30.68073],
                                [28.03905, 32.79057],
                                [30.46294, 32.79057],
                                [30.46294, 30.68073],
                                [28.03905, 30.68073],
                            ]
                        ],
                    },
                    "temporalCoverage": {
                        "endDate": "2020-02-10 00:00:00",
                        "startDate": "2020-02-10 00:00:00",
                    },
                    "publishDate": "2020-02-10 08:24:46",
                },
            ],
            "recordsReturned": 5,
            "totalHits": 139,
            "startingNumber": 6,
            "nextRecord": 11,
        }
    }

    DOWNLOAD_OPTION_RETURN = {
        "data": [
            {
                "id": "5e83d0b8e7f6734c",
                "entityId": "LC81780382020041LGN00",
                "available": True,
                "filesize": 9186067,
                "productName": "LandsatLook Natural Color Image",
                "downloadSystem": "dds",
            },
            {
                "entityId": "LC81780382020041LGN00",
                "available": False,
                "downloadSystem": "wms",
            },
        ]
    }

    def setUp(self):
        self.provider = "usgs"
        self.api_plugin = self.get_search_plugin(provider=self.provider)

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    def test_plugins_apis_usgs_authenticate(self, mock_api_logout, mock_api_login):
        """UsgsApi.authenticate must return try to login"""
        # no credential
        self.api_plugin.authenticate()
        mock_api_login.assert_called_once_with("", "", save=True)
        mock_api_logout.assert_not_called()
        mock_api_login.reset_mock()

        # with credentials
        self.api_plugin.config.credentials = {
            "username": "foo",
            "password": "bar",
        }
        self.api_plugin.authenticate()
        mock_api_login.assert_called_once_with("foo", "bar", save=True)
        mock_api_logout.assert_not_called()
        mock_api_login.reset_mock()

        # with USGSAuthExpiredError
        mock_api_login.side_effect = [USGSAuthExpiredError(), None]
        self.api_plugin.authenticate()
        self.assertEqual(mock_api_login.call_count, 2)
        self.assertEqual(mock_api_logout.call_count, 1)
        mock_api_login.reset_mock()
        mock_api_logout.reset_mock()

        # with obsolete `.usgs` API file (USGSError)
        mock_api_login.side_effect = [
            USGSError("USGS error"),
            None,
        ]
        with (
            mock.patch("os.remove", autospec=True) as mock_os_remove,
            mock.patch("os.path.isfile", autospec=True) as mock_isfile,
        ):
            self.api_plugin.authenticate()
            self.assertEqual(mock_api_login.call_count, 2)
            self.assertEqual(mock_api_logout.call_count, 0)
            mock_isfile.assert_called_once_with(USGS_TMPFILE)
            mock_os_remove.assert_called_once_with(USGS_TMPFILE)
        mock_api_login.reset_mock()
        mock_api_logout.reset_mock()

        # with invalid credentials / USGSError
        mock_api_login.side_effect = USGSError()
        with (
            mock.patch("os.remove", autospec=True),
            mock.patch("os.path.isfile", autospec=True),
            self.assertRaises(AuthenticationError),
        ):
            self.api_plugin.authenticate()
            self.assertEqual(mock_api_login.call_count, 2)
            mock_api_logout.assert_not_called()

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch(
        "usgs.api.scene_search",
        autospec=True,
        return_value=SCENE_SEARCH_RETURN,
    )
    @mock.patch(
        "usgs.api.download_options",
        autospec=True,
        return_value=DOWNLOAD_OPTION_RETURN,
    )
    def test_plugins_apis_usgs_query(
        self,
        mock_api_download_options,
        mock_api_scene_search,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.query must search using usgs api"""

        search_kwargs = {
            "productType": "LANDSAT_C2L1",
            "startTimeFromAscendingNode": "2020-02-01",
            "completionTimeFromAscendingNode": "2020-02-10",
            "geometry": get_geometry_from_various(geometry=[10, 20, 30, 40]),
            "prep": PreparedSearch(
                items_per_page=5,
                page=2,
            ),
        }
        search_results, total_count = self.api_plugin.query(**search_kwargs)
        mock_api_scene_search.assert_called_once_with(
            "landsat_ot_c2_l1",
            start_date="2020-02-01",
            end_date="2020-02-10",
            ll={"longitude": 10.0, "latitude": 20.0},
            ur={"longitude": 30.0, "latitude": 40.0},
            max_results=5,
            starting_number=6,
        )
        self.assertEqual(search_results[0].provider, "usgs")
        self.assertEqual(search_results[0].product_type, "LANDSAT_C2L1")
        self.assertEqual(
            search_results[0].geometry,
            shape(
                mock_api_scene_search.return_value["data"]["results"][0][
                    "spatialBounds"
                ]
            ),
        )
        self.assertEqual(
            search_results[0].properties["id"],
            mock_api_scene_search.return_value["data"]["results"][0]["displayId"],
        )
        self.assertEqual(
            search_results[0].properties["cloudCover"],
            float(
                mock_api_scene_search.return_value["data"]["results"][0]["cloudCover"]
            ),
        )
        self.assertEqual(search_results[0].properties["storageStatus"], ONLINE_STATUS)
        self.assertEqual(
            total_count, mock_api_scene_search.return_value["data"]["totalHits"]
        )

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch(
        "usgs.api.scene_search",
        autospec=True,
        return_value=SCENE_SEARCH_RETURN,
    )
    @mock.patch(
        "usgs.api.download_options",
        autospec=True,
        return_value=DOWNLOAD_OPTION_RETURN,
    )
    @mock.patch(
        "usgs.api.dataset_filters",
        autospec=True,
        return_value={
            "data": [
                {"id": "foo_id", "searchSql": "DONT_USE_THIS !"},
                {"id": "bar_id", "searchSql": "USE_THIS_ID !"},
            ]
        },
    )
    def test_plugins_apis_usgs_query_by_id(
        self,
        mock_dataset_filters,
        mock_api_download_options,
        mock_api_scene_search,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.query by id must search using usgs api"""

        search_kwargs = {
            "productType": "LANDSAT_C2L1",
            "id": "SOME_PRODUCT_ID",
            "prep": PreparedSearch(
                items_per_page=500,
                page=1,
            ),
        }
        search_results, total_count = self.api_plugin.query(**search_kwargs)
        mock_api_scene_search.assert_called_once_with(
            "landsat_ot_c2_l1",
            where={"filter_id": "bar_id", "value": "SOME_PRODUCT_ID"},
            start_date=None,
            end_date=None,
            ll=None,
            ur=None,
            max_results=500,
            starting_number=1,
        )
        self.assertEqual(search_results[0].provider, "usgs")
        self.assertEqual(search_results[0].product_type, "LANDSAT_C2L1")
        self.assertEqual(len(search_results), 1)

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch("usgs.api.download_request", autospec=True)
    def test_plugins_apis_usgs_download(
        self,
        mock_api_download_request,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.download must donwload using usgs api"""

        @responses.activate
        def run():
            product = EOProduct(
                "peps",
                dict(
                    geometry="POINT (0 0)",
                    title="dummy_product",
                    id="dummy",
                    entityId="dummyEntityId",
                    productId="dummyProductId",
                    productType="L8_OLI_TIRS_C1L1",
                ),
            )
            product.location = product.remote_location = product.properties[
                "downloadLink"
            ] = "http://somewhere"
            product.properties["id"] = "someproduct"

            responses.add(
                responses.GET,
                "http://path/to/product",
                status=200,
                content_type="application/octet-stream",
                body=b"something",
                auto_calculate_content_length=True,
                match=[
                    responses.matchers.request_kwargs_matcher(
                        dict(stream=True, timeout=DEFAULT_DOWNLOAD_WAIT * 60)
                    )
                ],
            )

            # missing download_request return value
            with self.assertRaises(NotAvailableError):
                self.api_plugin.download(product, output_dir=self.tmp_home_dir.name)

            # download 1 available product
            mock_api_download_request.return_value = {
                "data": {"preparingDownloads": [{"url": "http://path/to/product"}]}
            }

            path = self.api_plugin.download(product, output_dir=self.tmp_home_dir.name)

            self.assertEqual(len(responses.calls), 1)
            self.assertIn(
                list(USER_AGENT.items())[0], responses.calls[0].request.headers.items()
            )
            responses.calls.reset()

            self.assertEqual(
                path, os.path.join(self.tmp_home_dir.name, "dummy_product")
            )
            self.assertTrue(os.path.isfile(path))

            # check file extension
            # tar
            os.remove(path)
            with mock.patch("tarfile.is_tarfile", return_value=True, autospec=True):
                path = self.api_plugin.download(
                    product, output_dir=self.tmp_home_dir.name, extract=False
                )
                self.assertEqual(
                    path, os.path.join(self.tmp_home_dir.name, "dummy_product.tar.gz")
                )
            # zip
            os.remove(path)
            product.product_type = "S2_MSI_L1C"
            with mock.patch("zipfile.is_zipfile", return_value=True, autospec=True):
                path = self.api_plugin.download(
                    product, output_dir=self.tmp_home_dir.name, extract=False
                )
                self.assertEqual(
                    path, os.path.join(self.tmp_home_dir.name, "dummy_product.zip")
                )

        run()



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Annotated

if TYPE_CHECKING:
    from typing import Optional

    from typing_extensions import Doc, Self


class EodagError(Exception):
    """General EODAG error"""


class PluginNotFoundError(EodagError):
    """Error when looking for a plugin class that was not defined"""


class PluginImplementationError(EodagError):
    """Error when a plugin does not behave as expected"""


class MisconfiguredError(EodagError):
    """An error indicating a Search Plugin that is not well configured"""


class AddressNotFound(EodagError):
    """An error indicating the address of a subdataset was not found"""


class UnsupportedProvider(EodagError):
    """An error indicating that eodag does not support a provider"""


class UnsupportedProductType(EodagError):
    """An error indicating that eodag does not support a product type"""

    def __init__(self, product_type: str) -> None:
        self.product_type = product_type


class UnsupportedDatasetAddressScheme(EodagError):
    """An error indicating that eodag does not yet support an address scheme for
    accessing raster subdatasets"""


class AuthenticationError(EodagError):
    """An error indicating that an authentication plugin did not succeeded
    authenticating a user"""


class NotAvailableError(EodagError):
    """An error indicating that the product is not available for download"""


class NoMatchingProductType(EodagError):
    """An error indicating that eodag was unable to derive a product type from a set
    of search parameters"""


class STACOpenerError(EodagError):
    """An error indicating that a STAC file could not be opened"""


class InvalidDataError(EodagError):
    """Raised when data is invalid, malformed, or corrupt and cannot be processed as expected."""


class RequestError(EodagError):
    """An error indicating that a request has failed. Usually eodag functions
    and methods should catch and skip this"""

    status_code: Annotated[Optional[int], Doc("HTTP status code")] = None

    @classmethod
    def from_error(cls, error: Exception, msg: Optional[str] = None) -> Self:
        """Generate a RequestError from an Exception"""
        status_code = getattr(error, "code", None)
        text = getattr(error, "msg", None)

        response = getattr(error, "response", None)
        # Explicitly test for None because response objects are considered false if they
        # have a status code other than 200
        if response is not None:
            status_code = response.status_code
            text = " ".join([text or "", response.text])

        text = text or str(error)

        e = cls(msg, text) if msg else cls(text)
        e.status_code = status_code
        return e


class ValidationError(RequestError):
    """Error validating data"""

    def __init__(self, message: str, parameters: set[str] = set()) -> None:
        self.message = message
        self.parameters = parameters

    @classmethod
    def from_error(cls, error: Exception, msg: Optional[str] = None) -> Self:
        """Override parent from_error to handle ValidationError specificities."""
        setattr(error, "msg", msg)
        validation_error = super().from_error(error)
        return validation_error


class DownloadError(RequestError):
    """An error indicating something wrong with the download process"""


class TimeOutError(RequestError):
    """An error indicating that a timeout has occurred"""

    def __init__(
        self, exception: Optional[Exception] = None, timeout: Optional[float] = None
    ) -> None:
        url = getattr(getattr(exception, "request", None), "url", None)
        timeout_msg = f"({timeout}s)" if timeout else ""
        message = (
            f"Request timeout {timeout_msg} for URL {url}" if url else str(exception)
        )
        super().__init__(message)



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import os
import ssl
import sys
import unittest
from contextlib import closing
from datetime import datetime
from io import StringIO
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import mock

from requests.exceptions import RequestException

from tests.context import (
    HTTP_REQ_TIMEOUT,
    USER_AGENT,
    DownloadedCallback,
    ProgressCallback,
    RequestError,
    deepcopy,
    fetch_json,
    flatten_top_directories,
    get_bucket_name_and_prefix,
    get_ssl_context,
    get_timestamp,
    is_env_var_true,
    merge_mappings,
    path_to_uri,
    setup_logging,
    uri_to_path,
)


class TestUtils(unittest.TestCase):
    """Unit tests for utility functions in eodag.utils."""

    def setUp(self) -> None:
        """Set up logging before each test."""
        super(TestUtils, self).setUp()
        setup_logging(verbose=1)

    def tearDown(self) -> None:
        """Reset logging after each test."""
        super(TestUtils, self).tearDown()
        # reset logging
        logger = logging.getLogger("eodag")
        logger.handlers = []
        logger.level = 0

    def test_utils_get_timestamp(self):
        """Test get_timestamp returns correct UNIX timestamp for various date formats"""
        # Date to timestamp to date, this assumes the date is in UTC
        requested_date = "2020-08-08"  # Considered as 2020-08-08T00:00:00Z
        ts_in_secs = get_timestamp(requested_date)
        expected_dt = datetime.strptime(requested_date, "%Y-%m-%d")
        actual_utc_dt = datetime.utcfromtimestamp(ts_in_secs)
        self.assertEqual(actual_utc_dt, expected_dt)

        # Handle UTC datetime
        self.assertEqual(get_timestamp("2021-04-21T18:27:19.123Z"), 1619029639.123)
        # If date/datetime not in UTC, it assumes it's in UTC
        self.assertEqual(
            get_timestamp("2021-04-21T18:27:19.123"),
            get_timestamp("2021-04-21T18:27:19.123Z"),
        )
        self.assertEqual(
            get_timestamp("2021-04-21"), get_timestamp("2021-04-21T00:00:00.000Z")
        )

        # Non UTC datetime are also supported
        self.assertEqual(get_timestamp("2021-04-21T00:00:00+02:00"), 1618956000)

    def test_is_env_var_true(self):
        """Test is_env_var_true correctly interprets environment variable values."""
        test_cases = {
            "true": True,
            "True": True,
            "yes": True,
            "on": True,
            "1": True,
            "false": False,
            "False": False,
            "0": False,
            "no": False,
            "": False,
        }
        for value, expected in test_cases.items():
            with mock.patch.dict(os.environ, {"TEST_ENV_VAR": value}):
                self.assertEqual(is_env_var_true("TEST_ENV_VAR"), expected)

    def test_uri_to_path(self):
        """Test uri_to_path converts file URIs to file system paths."""
        if sys.platform == "win32":
            expected_path = r"C:\tmp\file.txt"
            tested_uri = r"file:///C:/tmp/file.txt"
            other_tested_uri = r"file:/C:/tmp/file.txt"
        else:
            expected_path = "/tmp/file.txt"
            tested_uri = "file:///tmp/file.txt"
            other_tested_uri = "file:/tmp/file.txt"
        actual_path = uri_to_path(tested_uri)
        self.assertEqual(actual_path, expected_path)
        actual_path = uri_to_path(other_tested_uri)
        self.assertEqual(actual_path, expected_path)
        with self.assertRaises(ValueError):
            uri_to_path("not_a_uri")

    def test_ssl_context(self):
        """Test get_ssl_context returns SSL context with correct settings."""

        ssl_ctx = get_ssl_context(False)
        self.assertEqual(ssl_ctx.verify_mode, ssl.CERT_NONE)
        self.assertEqual(ssl_ctx.check_hostname, False)

        ssl_ctx = get_ssl_context(True)
        self.assertEqual(ssl_ctx.verify_mode, ssl.CERT_REQUIRED)
        self.assertEqual(ssl_ctx.check_hostname, True)

    def test_path_to_uri(self):
        """Test path_to_uri converts file system paths to file URIs."""
        if sys.platform == "win32":
            self.assertEqual(path_to_uri(r"C:\tmp\file.txt"), "file:///C:/tmp/file.txt")
        else:
            self.assertEqual(path_to_uri("/tmp/file.txt"), "file:///tmp/file.txt")

    def test_downloaded_callback(self):
        """Test DownloadedCallback is callable and handles product parameter"""
        downloaded_callback = DownloadedCallback()
        self.assertTrue(callable(downloaded_callback))
        try:
            downloaded_callback(product=None)
        except TypeError as e:
            self.fail(f"DownloadedCallback got an error when called: {e}")

    def test_progresscallback_init(self):
        """Test ProgressCallback can be instantiated with default values"""
        with ProgressCallback() as bar:
            self.assertEqual(bar.unit, "B")
            self.assertEqual(bar.unit_scale, True)
            self.assertEqual(bar.desc, "")

    def test_progresscallback_init_customize(self):
        """Test ProgressCallback can be instantiated with custom values"""
        with ProgressCallback(unit="foo", unit_scale=False, desc="bar", total=5) as bar:
            self.assertEqual(bar.unit, "foo")
            self.assertEqual(bar.unit_scale, False)
            self.assertEqual(bar.desc, "bar")
            self.assertEqual(bar.total, 5)

    def test_progresscallback_copy(self):
        """Test ProgressCallback can be copied with the same attributes"""
        with ProgressCallback(unit="foo", unit_scale=False, desc="bar", total=5) as bar:
            with bar.copy() as another_bar:
                self.assertEqual(another_bar.unit, "foo")
                self.assertEqual(another_bar.unit_scale, False)
                self.assertEqual(another_bar.desc, "bar")
                self.assertEqual(another_bar.total, 5)

    def test_progresscallback_disable(self):
        """Test ProgressCallback can be disabled via logging or parameter"""
        # enabled
        with closing(StringIO()) as tqdm_out:
            with ProgressCallback(total=2, file=tqdm_out) as bar:
                bar(1)
            self.assertIn("50%", tqdm_out.getvalue())

        # disabled using setup_logging
        setup_logging(verbose=1, no_progress_bar=True)
        with closing(StringIO()) as tqdm_out:
            with ProgressCallback(total=2, file=tqdm_out) as bar:
                bar(1)
            self.assertEqual(tqdm_out.getvalue(), "")
        setup_logging(verbose=1, no_progress_bar=False)

        # disabled using tqdm parameter
        with closing(StringIO()) as tqdm_out:
            with ProgressCallback(total=2, file=tqdm_out, disable=True) as bar:
                bar(1)
            self.assertEqual(tqdm_out.getvalue(), "")

    def test_merge_mappings(self):
        """Test merge_mappings merges configuration mappings correctly."""

        # nested dict
        mapping = {"foo": {"keyA": "obsolete"}}
        merge_mappings(mapping, {"foo": {"keya": "new"}})
        self.assertEqual(mapping, {"foo": {"keyA": "new"}})

        # list replaced by string
        mapping = {"keyA": ["obsolete1", "obsolete2"]}
        merge_mappings(mapping, {"keya": "new"})
        self.assertEqual(mapping, {"keyA": "new"})

        # string replaced by list
        mapping = {"keyA": "obsolete"}
        merge_mappings(mapping, {"keya": ["new1", "new2"]})
        self.assertEqual(mapping, {"keyA": ["new1", "new2"]})

        # bool replaced by str
        mapping = {"keyA": True}
        merge_mappings(mapping, {"keya": "fAlSe"})
        self.assertEqual(mapping, {"keyA": False})

        # int replaced by str
        mapping = {"keyA": 1}
        merge_mappings(mapping, {"keya": "2"})
        self.assertEqual(mapping, {"keyA": 2})

        # override ignored if cast cannot be perfomed to origin type
        mapping = {"keyA": True}
        merge_mappings(mapping, {"keya": "bar"})
        self.assertEqual(mapping, {"keyA": True})

    def test_get_bucket_name_and_prefix(self):
        """Test get_bucket_name_and_prefix extracts bucket and prefix from URLs"""
        self.assertEqual(
            get_bucket_name_and_prefix(
                "s3://sentinel-s2-l1c/tiles/50/R/LR/2021/6/8/0/B02.jp2"
            ),
            ("sentinel-s2-l1c", "tiles/50/R/LR/2021/6/8/0/B02.jp2"),
        )
        self.assertEqual(
            get_bucket_name_and_prefix(
                "https://sentinel-cogs.s3.us-west-2.amazonaws.com/sentinel-s2-l2a-cogs/44/2022/10/S2B_20221011/B01.tif"
            ),
            (
                "sentinel-cogs",
                "sentinel-s2-l2a-cogs/44/2022/10/S2B_20221011/B01.tif",
            ),
        )
        self.assertEqual(
            get_bucket_name_and_prefix(
                "https://s3.us-west-2.amazonaws.com/sentinel-cogs/sentinel-s2-l2a-cogs/44/2022/10/S2B_20221011/B01.tif"
            ),
            (
                "sentinel-cogs",
                "sentinel-s2-l2a-cogs/44/2022/10/S2B_20221011/B01.tif",
            ),
        )
        self.assertEqual(
            get_bucket_name_and_prefix(
                "https://obs.eu-de.otc.t-systems.com/s2-l1c-2021-q4/30/T/2021/11/14/S2A_MSIL1C_20211114T110321_N0301",
                0,
            ),
            ("s2-l1c-2021-q4", "30/T/2021/11/14/S2A_MSIL1C_20211114T110321_N0301"),
        )
        self.assertEqual(
            get_bucket_name_and_prefix(
                "products/2021/1/5/S2A_MSIL1C_20210105T105441_N0209_R051_T30TYN_20210105T130129",
            ),
            (
                None,
                "products/2021/1/5/S2A_MSIL1C_20210105T105441_N0209_R051_T30TYN_20210105T130129",
            ),
        )
        self.assertEqual(
            get_bucket_name_and_prefix(
                "bucket/path/to/product",
                0,
            ),
            ("bucket", "path/to/product"),
        )
        self.assertEqual(
            get_bucket_name_and_prefix(
                "http://foo/bar/bucket/path/to/product",
                1,
            ),
            ("bucket", "path/to/product"),
        )

    def test_flatten_top_dirs(self):
        """Test flatten_top_directories flattens nested directory structures"""
        with TemporaryDirectory() as nested_dir_root:
            os.makedirs(os.path.join(nested_dir_root, "a", "b", "c1"))
            os.makedirs(os.path.join(nested_dir_root, "a", "b", "c2"))
            # create empty files
            open(os.path.join(nested_dir_root, "a", "b", "c1", "foo"), "a").close()
            open(os.path.join(nested_dir_root, "a", "b", "c2", "bar"), "a").close()
            open(os.path.join(nested_dir_root, "a", "b", "c2", "baz"), "a").close()

            flatten_top_directories(nested_dir_root)

            dir_content = list(Path(nested_dir_root).glob("**/*"))

            self.assertEqual(len(dir_content), 5)
            self.assertIn(Path(nested_dir_root) / "c1", dir_content)
            self.assertIn(Path(nested_dir_root) / "c1" / "foo", dir_content)
            self.assertIn(Path(nested_dir_root) / "c2", dir_content)
            self.assertIn(Path(nested_dir_root) / "c2" / "bar", dir_content)
            self.assertIn(Path(nested_dir_root) / "c2" / "baz", dir_content)

    def test_flatten_top_dirs_single_file(self):
        """Test flatten_top_directories flattens structure with a single file"""
        with TemporaryDirectory() as nested_dir_root:
            os.makedirs(os.path.join(nested_dir_root, "a", "b", "c1"))
            # create empty file
            open(os.path.join(nested_dir_root, "a", "b", "c1", "foo"), "a").close()

            flatten_top_directories(nested_dir_root)

            dir_content = list(Path(nested_dir_root).glob("**/*"))

            self.assertEqual(len(dir_content), 1)
            self.assertIn(Path(nested_dir_root) / "foo", dir_content)

    def test_flatten_top_dirs_given_subdir(self):
        """Test flatten_top_directories flattens structure from a given subdirectory"""
        with TemporaryDirectory() as nested_dir_root:
            os.makedirs(os.path.join(nested_dir_root, "a", "b", "c1"))
            os.makedirs(os.path.join(nested_dir_root, "a", "b", "c2"))
            # create empty files
            open(os.path.join(nested_dir_root, "a", "b", "c1", "foo"), "a").close()
            open(os.path.join(nested_dir_root, "a", "b", "c2", "bar"), "a").close()
            open(os.path.join(nested_dir_root, "a", "b", "c2", "baz"), "a").close()

            flatten_top_directories(nested_dir_root, os.path.join(nested_dir_root, "a"))

            dir_content = list(Path(nested_dir_root).glob("**/*"))

            self.assertEqual(len(dir_content), 6)
            self.assertIn(Path(nested_dir_root) / "b", dir_content)
            self.assertIn(Path(nested_dir_root) / "b" / "c1", dir_content)
            self.assertIn(Path(nested_dir_root) / "b" / "c1" / "foo", dir_content)
            self.assertIn(Path(nested_dir_root) / "b" / "c2", dir_content)
            self.assertIn(Path(nested_dir_root) / "b" / "c2" / "bar", dir_content)
            self.assertIn(Path(nested_dir_root) / "b" / "c2" / "baz", dir_content)

    def test_deepcopy(self):
        """Test deepcopy performs a recursive copy of objects"""
        original = {"a": [{"b": [0, 1, 2]}]}
        shallow_copied = copy.copy(original)
        deep_copied = deepcopy(original)
        # change original
        original["a"][0]["b"][0] = 5
        # shallow copy also changed
        self.assertEqual(shallow_copied["a"][0]["b"][0], 5)
        # deep copy did not change
        self.assertEqual(deep_copied["a"][0]["b"][0], 0)

    def test_fetch_json(self):
        """Test fetch_json fetches and parses JSON from a URL, handles errors"""

        # distant
        file_url = "https://foo.bar"
        with unittest.mock.patch(
            "eodag.utils.requests.requests.sessions.Session.get",
            autospec=True,
        ) as mock_get:
            mock_get.return_value = unittest.mock.Mock()
            mock_get.return_value.json.return_value = {"foo": "bar"}
            file_content = fetch_json(file_url)
            self.assertEqual(file_content["foo"], "bar")
            mock_get.assert_called_once_with(
                mock.ANY,
                file_url,
                headers=USER_AGENT,
                auth=None,
                timeout=HTTP_REQ_TIMEOUT,
            )

        # distant error
        with unittest.mock.patch(
            "eodag.utils.requests.requests.get",
            autospec=True,
            side_effect=RequestException,
        ) as mock_get:
            self.assertRaises(RequestError, fetch_json, file_url)



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import unittest
from pathlib import Path
from shutil import copyfile
from tempfile import TemporaryDirectory

import yaml

from tests.context import TEST_RESOURCES_PATH, AwsDownload, EOProduct, NotAvailableError


class TestSafeBuild(unittest.TestCase):
    def setUp(self):
        super(TestSafeBuild, self).setUp()

        self.awsd = AwsDownload("some_provider", {})
        self.logger = logging.getLogger("eodag.download.aws")

        self.tmp_download_dir = TemporaryDirectory()
        self.tmp_download_path = self.tmp_download_dir.name

        with open(
            os.path.join(TEST_RESOURCES_PATH, "safe_build", "aws_sentinel_chunks.yml"),
            "r",
        ) as fh:
            self.aws_sentinel_chunks = yaml.load(fh, Loader=yaml.SafeLoader)

    def tearDown(self):
        self.tmp_download_dir.cleanup()

    def test_safe_build_out_of_pattern(self):
        """Cannot build SAFE product from out of pattern file"""
        prod = EOProduct(
            provider="some_provider",
            properties=self.aws_sentinel_chunks["S1_SAR_GRD"]["properties"],
            productType="S1_SAR_GRD",
        )

        def chunk():
            return None

        chunk.key = "path/to/some/dummy_chunk"

        with self.assertRaisesRegex(
            NotAvailableError, f"Ignored {chunk.key} out of SAFE matching pattern"
        ):
            self.awsd.get_chunk_dest_path(
                prod,
                chunk,
                build_safe=True,
            )

    def test_safe_build_no_safe(self):
        """Do not use patterns if no SAFE is asked"""
        prod = EOProduct(
            provider="some_provider",
            properties=self.aws_sentinel_chunks["S1_SAR_GRD"]["properties"],
            productType="S1_SAR_GRD",
        )

        def chunk():
            return None

        chunk.key = "path/to/some/dummy_chunk"
        chunk_dest_path = self.awsd.get_chunk_dest_path(
            prod,
            chunk,
            build_safe=False,
        )
        self.assertEqual(chunk_dest_path, chunk.key)

        chunk_dest_path = self.awsd.get_chunk_dest_path(
            prod, chunk, build_safe=False, dir_prefix="path/to"
        )
        self.assertEqual(chunk_dest_path, "some/dummy_chunk")

    def test_safe_build_s1_sar_grd(self):
        """build S1_SAR_GRD SAFE product with empty files and check content"""

        prod = EOProduct(
            provider="some_provider",
            properties=self.aws_sentinel_chunks["S1_SAR_GRD"]["properties"],
            productType="S1_SAR_GRD",
        )

        product_path = os.path.join(self.tmp_download_path, prod.properties["title"])

        def chunk():
            return None

        for chunk_key in self.aws_sentinel_chunks["S1_SAR_GRD"]["chunks"]:
            chunk.key = chunk_key
            chunk_dest_rel_path = self.awsd.get_chunk_dest_path(
                prod,
                chunk,
                build_safe=True,
            )

            chunk_abs_path = os.path.join(product_path, chunk_dest_rel_path)
            chunk_abs_path_dir = os.path.dirname(chunk_abs_path)
            if not os.path.isdir(chunk_abs_path_dir):
                os.makedirs(chunk_abs_path_dir)

            Path(chunk_abs_path).touch()

        copyfile(
            os.path.join(TEST_RESOURCES_PATH, "safe_build", "manifest.safe.S1_SAR_GRD"),
            os.path.join(product_path, "manifest.safe"),
        )

        with self.assertLogs(self.logger, logging.WARN) as cm:
            # assertLogs fails if no warning is raised
            self.logger.warning("Dummy warning")
            self.awsd.check_manifest_file_list(product_path)

        self.assertEqual(len(cm.output), 1)
        self.assertIn("Dummy warning", cm.output[0])

    def test_safe_build_s2_msi_l2a(self):
        """build S2_MSI_L2A SAFE product with empty files and check content"""

        prod = EOProduct(
            provider="some_provider",
            properties=self.aws_sentinel_chunks["S2_MSI_L2A"]["properties"],
            productType="S2_MSI_L2A",
        )

        product_path = os.path.join(self.tmp_download_path, prod.properties["title"])

        def chunk():
            return None

        for chunk_key in self.aws_sentinel_chunks["S2_MSI_L2A"]["chunks"]:
            chunk.key = chunk_key
            chunk_dest_rel_path = self.awsd.get_chunk_dest_path(
                prod,
                chunk,
                build_safe=True,
            )

            chunk_abs_path = os.path.join(product_path, chunk_dest_rel_path)
            chunk_abs_path_dir = os.path.dirname(chunk_abs_path)
            if not os.path.isdir(chunk_abs_path_dir):
                os.makedirs(chunk_abs_path_dir)

            Path(chunk_abs_path).touch()

        copyfile(
            os.path.join(TEST_RESOURCES_PATH, "safe_build", "manifest.safe.S2_MSI_L2A"),
            os.path.join(product_path, "manifest.safe"),
        )

        self.awsd.finalize_s2_safe_product(product_path)

        with self.assertLogs(self.logger, logging.WARN) as cm:
            # assertLogs fails if no warning is raised
            self.logger.warning("Dummy warning")
            self.awsd.check_manifest_file_list(product_path)

        self.assertEqual(len(cm.output), 1)
        self.assertIn("Dummy warning", cm.output[0])

    def test_safe_build_s2_msi_l1c(self):
        """build S2_MSI_L1C SAFE product with empty files and check content"""

        prod = EOProduct(
            provider="some_provider",
            properties=self.aws_sentinel_chunks["S2_MSI_L1C"]["properties"],
            productType="S2_MSI_L1C",
        )

        product_path = os.path.join(self.tmp_download_path, prod.properties["title"])

        def chunk():
            return None

        for chunk_key in self.aws_sentinel_chunks["S2_MSI_L1C"]["chunks"]:
            chunk.key = chunk_key
            chunk_dest_rel_path = self.awsd.get_chunk_dest_path(
                prod,
                chunk,
                build_safe=True,
            )

            chunk_abs_path = os.path.join(product_path, chunk_dest_rel_path)
            chunk_abs_path_dir = os.path.dirname(chunk_abs_path)
            if not os.path.isdir(chunk_abs_path_dir):
                os.makedirs(chunk_abs_path_dir)

            Path(chunk_abs_path).touch()

        copyfile(
            os.path.join(TEST_RESOURCES_PATH, "safe_build", "manifest.safe.S2_MSI_L1C"),
            os.path.join(product_path, "manifest.safe"),
        )

        self.awsd.finalize_s2_safe_product(product_path)

        with self.assertLogs(self.logger, logging.WARN) as cm:
            # assertLogs fails if no warning is raised
            self.logger.warning("Dummy warning")
            self.awsd.check_manifest_file_list(product_path)

        self.assertEqual(len(cm.output), 2)
        self.assertIn("Dummy warning", cm.output[0])
        # known missing file, see https://github.com/CS-SI/eodag/pull/218#issuecomment-816770353
        self.assertIn("PVI.jp2 is missing", cm.output[1])



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import unittest

from tests import TEST_RESOURCES_PATH
from tests.context import STACOpenerError, _TextOpener, fetch_stac_items


class TestStacReader(unittest.TestCase):
    def setUp(self):
        super(TestStacReader, self).setUp()

        self.cat_dir_path = self.root_cat = os.path.join(TEST_RESOURCES_PATH, "stac")
        self.root_cat = os.path.join(self.cat_dir_path, "catalog.json")
        self.root_cat_len = 5
        self.child_cat = os.path.join(
            self.cat_dir_path, "country", "FRA", "year", "2018", "2018.json"
        )
        self.child_cat_len = 2
        self.item = os.path.join(
            os.path.dirname(self.child_cat),
            "items",
            "S2A_MSIL1C_20181231T141041_N0207_R110_T21NYF_20181231T155050",
            "S2A_MSIL1C_20181231T141041_N0207_R110_T21NYF_20181231T155050.json",
        )
        self.singlefile_cat = os.path.join(TEST_RESOURCES_PATH, "stac_singlefile.json")
        self.singlefile_cat_len = 5

    def test_stac_reader_fetch_child(self):
        """fetch_stac_items from child catalog must provide items"""
        items = fetch_stac_items(self.child_cat)
        self.assertIsInstance(items, list)
        self.assertEqual(len(items), self.child_cat_len)
        self.assertEqual(items[0]["type"], "Feature")
        self.assertEqual(items[0]["collection"], "S2_MSI_L1C")

    def test_stac_reader_fetch_root_not_recursive(self):
        """fetch_stac_items from root must provide an empty list when no recursive"""
        items = fetch_stac_items(self.root_cat, recursive=False)
        self.assertListEqual(items, [])

    def test_stac_reader_fetch_root_recursive(self):
        """fetch_stac_items from root must provide items when recursive"""
        items = fetch_stac_items(self.root_cat, recursive=True)
        self.assertEqual(len(items), self.root_cat_len)
        for item in items:
            self.assertEqual(item["type"], "Feature")
            self.assertEqual(item["collection"], "S2_MSI_L1C")

    def test_stac_reader_fetch_item(self):
        """fetch_stac_items from an item must return it"""
        item = fetch_stac_items(self.item)
        self.assertIsInstance(item, list)
        self.assertEqual(len(item), 1)
        self.assertEqual(item[0]["type"], "Feature")
        self.assertEqual(item[0]["collection"], "S2_MSI_L1C")

    @unittest.skip(
        "skipped as single-file-stac has been removed and is being rethought"
    )
    def test_stact_reader_fetch_singlefile_catalog(self):
        """fetch_stact_items must return all the items from a single file catalog"""
        items = fetch_stac_items(self.singlefile_cat)
        self.assertIsInstance(items, list)
        self.assertEqual(len(items), self.singlefile_cat_len)
        self.assertEqual(items[0]["type"], "Feature")

    def test_stact_reader_invalid_local_json(self):
        """read_local_json must raise a STACOpenerError on invalid local file path"""
        self.assertRaises(
            STACOpenerError,
            _TextOpener.read_local_json,
            "http://data.example.org/",
            True,
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
import re
import string
from datetime import datetime, timedelta, timezone
from random import SystemRandom
from typing import TYPE_CHECKING, Any, Optional
from urllib.parse import parse_qs, urlparse

import jwt
import requests
from lxml import etree
from requests.auth import AuthBase

from eodag.plugins.authentication import Authentication
from eodag.utils import (
    DEFAULT_TOKEN_EXPIRATION_MARGIN,
    HTTP_REQ_TIMEOUT,
    USER_AGENT,
    repeatfunc,
)
from eodag.utils.exceptions import (
    AuthenticationError,
    MisconfiguredError,
    RequestError,
    TimeOutError,
)

if TYPE_CHECKING:
    from requests import PreparedRequest, Response

    from eodag.config import PluginConfig


logger = logging.getLogger("eodag.auth.openid_connect")


class OIDCRefreshTokenBase(Authentication):
    """OIDC refresh token base class, to be used through specific OIDC flows plugins;
    Common mechanism to handle refresh token from all OIDC auth plugins;

    Plugins inheriting from this base class must implement the methods ``_request_new_token()`` and
    ``_get_token_with_refresh_token()``. Depending on the implementation of these methods they can have
    different configuration parameters.

    """

    jwks_client: jwt.PyJWKClient

    access_token: str
    access_token_expiration: datetime

    refresh_token: str
    refresh_token_expiration: datetime

    token_endpoint: str
    authorization_endpoint: str

    def __init__(self, provider: str, config: PluginConfig) -> None:
        super(OIDCRefreshTokenBase, self).__init__(provider, config)
        self.session = requests.Session()

        self.access_token = ""
        self.access_token_expiration = datetime.min.replace(tzinfo=timezone.utc)

        self.refresh_token = ""
        self.refresh_token_expiration = datetime.min.replace(tzinfo=timezone.utc)

        try:
            response = requests.get(self.config.oidc_config_url)
            response.raise_for_status()
            auth_config = response.json()
        except requests.HTTPError as e:
            raise MisconfiguredError(
                f"Cannot obtain OIDC endpoints from {self.config.oidc_config_url}"
                f"Request returned {e.response.text}."
            )

        self.jwks_client = jwt.PyJWKClient(auth_config["jwks_uri"])
        self.token_endpoint = auth_config["token_endpoint"]
        self.authorization_endpoint = auth_config["authorization_endpoint"]
        self.algorithms = auth_config["id_token_signing_alg_values_supported"]

    def decode_jwt_token(self, token: str) -> dict[str, Any]:
        """Decode JWT token."""
        try:
            key = self.jwks_client.get_signing_key_from_jwt(token).key
            if getattr(self.config, "allowed_audiences", None):
                return jwt.decode(
                    token,
                    key,
                    algorithms=self.algorithms,
                    # NOTE: Audience validation MUST match audience claim if set in token
                    # (https://pyjwt.readthedocs.io/en/stable/changelog.html?highlight=audience#id40)
                    audience=self.config.allowed_audiences,
                )
            else:
                return jwt.decode(
                    token,
                    key,
                    algorithms=self.algorithms,
                )
        except (jwt.exceptions.InvalidTokenError, jwt.exceptions.DecodeError) as e:
            raise AuthenticationError(e)

    def _get_access_token(self) -> str:
        now = datetime.now(timezone.utc)
        expiration_margin = timedelta(
            seconds=getattr(
                self.config, "token_expiration_margin", DEFAULT_TOKEN_EXPIRATION_MARGIN
            )
        )

        if self.access_token and self.access_token_expiration - now > expiration_margin:
            logger.debug(
                f"Existing access_token is still valid until {self.access_token_expiration.isoformat()}."
            )
            return self.access_token

        elif (
            self.refresh_token
            and self.refresh_token_expiration - now > expiration_margin
        ):
            response = self._get_token_with_refresh_token()
            logger.debug(
                "access_token expired, fetching new access_token using refresh_token"
            )
        else:
            logger.debug("access_token expired or not available yet, new token request")
            response = self._request_new_token()

        self.access_token = response[getattr(self.config, "token_key", "access_token")]
        self.access_token_expiration = datetime.fromtimestamp(
            self.decode_jwt_token(self.access_token)["exp"], timezone.utc
        )
        self.refresh_token = response.get(
            getattr(self.config, "refresh_token_key", "refresh_token"), ""
        )
        if self.refresh_token and response.get("refresh_expires_in", "0"):
            self.refresh_token_expiration = now + timedelta(
                seconds=int(response["refresh_expires_in"])
            )
        else:
            # refresh token does not expire but will be changed at each request
            self.refresh_token_expiration = now + timedelta(days=1000)

        return self.access_token

    def _request_new_token(self) -> dict[str, str]:
        """Fetch the access token with a new authentication"""
        raise NotImplementedError(
            "Incomplete OIDC refresh token retrieval mechanism implementation"
        )

    def _request_new_token_error(self, e: requests.RequestException) -> dict[str, str]:
        """Handle RequestException raised by `self._request_new_token()`"""
        if self.access_token:
            # try using already retrieved token if authenticate() fails (OTP use-case)
            return {
                "access_token": self.access_token,
                "expires_in": self.access_token_expiration.isoformat(),
            }
        response_text = getattr(e.response, "text", "").strip()
        # check if error is identified as auth_error in provider conf
        auth_errors = getattr(self.config, "auth_error_code", [None])
        if not isinstance(auth_errors, list):
            auth_errors = [auth_errors]
        if (
            e.response
            and hasattr(e.response, "status_code")
            and e.response.status_code in auth_errors
        ):
            raise AuthenticationError(
                f"Please check your credentials for {self.provider}.",
                f"HTTP Error {e.response.status_code} returned.",
                response_text,
            )
        # other error
        else:
            import traceback as tb

            logger.error(
                f"Provider {self.provider} returned {getattr(e.response, 'status_code', '')}: {response_text}"
            )
            raise AuthenticationError(
                "Something went wrong while trying to get access token:\n{}".format(
                    tb.format_exc()
                )
            )

    def _get_token_with_refresh_token(self) -> dict[str, str]:
        """Fetch the access token with the refresh token"""
        raise NotImplementedError(
            "Incomplete OIDC refresh token retrieval mechanism implementation"
        )


class OIDCAuthorizationCodeFlowAuth(OIDCRefreshTokenBase):
    """Implement the authorization code flow of the OpenIDConnect authorization specification.

    The `OpenID Connect <http://openid.net/specs/openid-connect-core-1_0.html>`_ specification
    adds an authentication layer on top of oauth 2.0. This plugin implements the
    `authorization code flow <http://openid.net/specs/openid-connect-core-1_0.html#Authentication>`_
    option of this specification.

    The particularity of this plugin is that it proceeds to a headless (not involving the user)
    interaction with the OpenID provider (if necessary) to authenticate a
    registered user with its username and password on the server and then granting to eodag the
    necessary rights. It does that using the client ID of the eodag provider that use it.
    If the client secret of the eodag provider using this plugin is known, it is used in conjunction
    with the client ID to do a BASIC Auth during the token exchange request.
    The headless interaction is fully configurable, and rely on XPATH to retrieve all the necessary
    information.

    :param provider: provider name
    :param config: Authentication plugin configuration:

        * :attr:`~eodag.config.PluginConfig.type` (``str``) (**mandatory**): OIDCAuthorizationCodeFlowAuth
        * :attr:`~eodag.config.PluginConfig.redirect_uri` (``str``) (**mandatory**):  The callback
          url that will handle the code given by the OIDC provider
        * :attr:`~eodag.config.PluginConfig.oidc_config_url` (``str``) (**mandatory**):
          The url to get the OIDC Provider's endpoints
        * :attr:`~eodag.config.PluginConfig.client_id` (``str``) (**mandatory**): The OIDC provider's
          client ID of the eodag provider
        * :attr:`~eodag.config.PluginConfig.user_consent_needed` (``bool``) (mandatory): Whether
          a user consent is needed during the authentication
        * :attr:`~eodag.config.PluginConfig.token_exchange_post_data_method` (``str``) (**mandatory**):
          One of: ``json``, ``data`` or ``params``. This is the way to pass the data to the
          POST request that is made to the token server. They correspond to the recognised keywords
          arguments of the Python `requests <http://docs.python-requests.org/>`_ library
        * :attr:`~eodag.config.PluginConfig.token_key` (``str``): The key pointing
          to the token in the json response to the POST request to the token server
        * :attr:`~eodag.config.PluginConfig.token_provision` (``str``) (**mandatory**): One of
          ``qs`` or ``header``. This is how the token obtained will be used to authenticate the
          user on protected requests. If ``qs`` is chosen, then ``token_qs_key`` is mandatory
        * :attr:`~eodag.config.PluginConfig.login_form_xpath` (``str``) (**mandatory**): The
          xpath to the HTML form element representing the user login form
        * :attr:`~eodag.config.PluginConfig.authentication_uri_source` (``str``) (**mandatory**): Where
          to look for the authentication_uri. One of ``config`` (in the configuration) or ``login-form``
          (use the 'action' URL found in the login form retrieved with login_form_xpath). If the
          value is ``config``, authentication_uri config param is mandatory
        * :attr:`~eodag.config.PluginConfig.authentication_uri` (``str``): (**mandatory if
          authentication_uri_source=config**) The URL of the authentication backend of the OIDC provider
        * :attr:`~eodag.config.PluginConfig.user_consent_form_xpath` (``str``): The xpath to
          the user consent form. The form is searched in the content of the response to the authorization request
        * :attr:`~eodag.config.PluginConfig.user_consent_form_data` (``dict[str, str]``): The data that
          will be passed with the POST request on the form 'action' URL. The data are given as
          key value pairs, the keys representing the data key and the value being either a 'constant'
          string value, or a string of the form 'xpath(<path-to-a-value-to-be-retrieved>)' and representing a
          value to be retrieved in the user consent form. The xpath must resolve directly to a
          string value, not to an HTML element. Example: ``xpath(//input[@name="sessionDataKeyConsent"]/@value)``
        * :attr:`~eodag.config.PluginConfig.additional_login_form_data` (``dict[str, str]``): A mapping
          giving additional data to be passed to the login POST request. The value follows
          the same rules as with user_consent_form_data
        * :attr:`~eodag.config.PluginConfig.exchange_url_error_pattern` (``dict[str, str]``): Key/value
          pairs of patterns/messages. If exchange_url contains the given pattern, the associated
          message will be sent in an AuthenticationError
        * :attr:`~eodag.config.PluginConfig.client_secret` (``str``): The OIDC provider's client
          secret of the eodag provider
        * :attr:`~eodag.config.PluginConfig.token_exchange_params` (``dict[str, str]``): mandatory
          keys for the dict: redirect_uri, client_id; A mapping between OIDC url query string
          and token handler query string params (only necessary if they are not the same as for OIDC).
          This is eodag provider dependant
        * :attr:`~eodag.config.PluginConfig.token_qs_key` (``str``): (mandatory when token_provision=qs)
          Refers to the name of the query param to be used in the query request
        * :attr:`~eodag.config.PluginConfig.refresh_token_key` (``str``): The key pointing to
          the refresh_token in the json response to the POST request to the token server
        * :attr:`~eodag.config.PluginConfig.token_expiration_margin` (``int``): The margin of time (in seconds)
          before a token is considered expired. Default: 60 seconds.

    """

    SCOPE = "openid"
    RESPONSE_TYPE = "code"
    CONFIG_XPATH_REGEX = re.compile(r"^xpath\((?P<xpath_value>.+)\)$")

    def __init__(self, provider: str, config: PluginConfig) -> None:
        super(OIDCAuthorizationCodeFlowAuth, self).__init__(provider, config)

    def validate_config_credentials(self) -> None:
        """Validate configured credentials"""
        super(OIDCAuthorizationCodeFlowAuth, self).validate_config_credentials()
        if getattr(self.config, "token_provision", None) not in ("qs", "header"):
            raise MisconfiguredError(
                'Provider config parameter "token_provision" must be one of "qs" or "header"'
            )
        if self.config.token_provision == "qs" and not getattr(
            self.config, "token_qs_key", ""
        ):
            raise MisconfiguredError(
                'Provider config parameter "token_provision" with value "qs" must have '
                '"token_qs_key" config parameter as well'
            )

    def authenticate(self) -> CodeAuthorizedAuth:
        """Authenticate"""
        self._get_access_token()

        return CodeAuthorizedAuth(
            self.access_token,
            self.config.token_provision,
            key=getattr(self.config, "token_qs_key", None),
        )

    def _request_new_token(self) -> dict[str, str]:
        """Fetch the access token with a new authentication"""
        logger.debug("Fetching access token from %s", self.token_endpoint)
        state = self.compute_state()
        authentication_response = self.authenticate_user(state)
        exchange_url = authentication_response.url

        for err_pattern, err_message in getattr(
            self.config, "exchange_url_error_pattern", {}
        ).items():
            if err_pattern in exchange_url:
                raise AuthenticationError(err_message)

        if not exchange_url.startswith(self.config.redirect_uri):
            if "Invalid username or password" in authentication_response.text:
                raise AuthenticationError("Invalid username or password")

            raise AuthenticationError(
                f"Could not authenticate user with provider {self.provider}.",
                "Please verify your credentials",
            )
        if self.config.user_consent_needed:
            user_consent_response = self.grant_user_consent(authentication_response)
            exchange_url = user_consent_response.url
        try:
            token_response = self.exchange_code_for_token(exchange_url, state)
            token_response.raise_for_status()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
        except requests.RequestException as e:
            return self._request_new_token_error(e)
        return token_response.json()

    def _get_token_with_refresh_token(self) -> dict[str, str]:
        """Fetch the access token with the refresh token"""
        logger.debug(
            "Fetching access token with refresh token from %s.", self.token_endpoint
        )
        token_data: dict[str, Any] = {
            "refresh_token": self.refresh_token,
            "grant_type": "refresh_token",
        }
        token_data = self._prepare_token_post_data(token_data)
        post_request_kwargs: Any = {
            self.config.token_exchange_post_data_method: token_data
        }
        ssl_verify = getattr(self.config, "ssl_verify", True)
        try:
            token_response = self.session.post(
                self.token_endpoint,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
                **post_request_kwargs,
            )
            token_response.raise_for_status()
        except requests.exceptions.Timeout as exc:
            raise TimeOutError(exc, timeout=HTTP_REQ_TIMEOUT) from exc
        except requests.RequestException as exc:
            logger.error(
                "Could not fetch access token with refresh token, executing new token request, error: %s",
                getattr(exc.response, "text", ""),
            )
            return self._request_new_token()
        return token_response.json()

    def authenticate_user(self, state: str) -> Response:
        """Authenticate user"""
        self.validate_config_credentials()
        params = {
            "client_id": self.config.client_id,
            "response_type": self.RESPONSE_TYPE,
            "scope": self.SCOPE,
            "state": state,
            "redirect_uri": self.config.redirect_uri,
        }
        ssl_verify = getattr(self.config, "ssl_verify", True)
        try:
            authorization_response = self.session.get(
                self.authorization_endpoint,
                params=params,
                headers=USER_AGENT,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
            )
        except requests.exceptions.Timeout as exc:
            raise TimeoutError(exc, "The authentication request timed out.") from exc
        except requests.RequestException as exc:
            raise RequestError.from_error(
                exc, "An error occurred while authenticating the user."
            ) from exc

        login_document = etree.HTML(authorization_response.text)
        login_forms = login_document.xpath(self.config.login_form_xpath)

        if not login_forms:
            # we assume user is already logged in
            # no form found because we got redirected to the redirect_uri
            return authorization_response

        login_form = login_forms[0]

        # Get the form data to pass to the login form from config or from the login form
        login_data = {
            key: self._constant_or_xpath_extracted(value, login_form)
            for key, value in getattr(
                self.config, "additional_login_form_data", {}
            ).items()
        }
        # Add the credentials
        login_data.update(self.config.credentials)

        # Retrieve the authentication_uri from the login form if so configured
        if self.config.authentication_uri_source == "login-form":
            # Given that the login_form_xpath resolves to an HTML element, if suffices to add '/@action' to get
            # the value of its action attribute to this xpath
            auth_uri = login_form.xpath(
                self.config.login_form_xpath.rstrip("/") + "/@action"
            )
            if not auth_uri or not auth_uri[0]:
                raise MisconfiguredError(
                    f"Could not get auth_uri from {self.config.login_form_xpath}"
                )
            auth_uri = auth_uri[0]
        else:
            auth_uri = getattr(self.config, "authentication_uri", None)
            if not auth_uri:
                raise MisconfiguredError("authentication_uri is missing")
        try:
            return self.session.post(
                auth_uri,
                data=login_data,
                headers=USER_AGENT,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
            )
        except requests.exceptions.Timeout as exc:
            raise TimeoutError(exc, "The authentication request timed out.") from exc
        except requests.RequestException as exc:
            raise RequestError.from_error(
                exc, "An error occurred while authenticating the user."
            ) from exc

    def grant_user_consent(self, authentication_response: Response) -> Response:
        """Grant user consent"""
        user_consent_document = etree.HTML(authentication_response.text)
        user_consent_form = user_consent_document.xpath(
            self.config.user_consent_form_xpath
        )[0]
        # Get the form data to pass to the consent form from config or from the consent form
        user_consent_data = {
            key: self._constant_or_xpath_extracted(value, user_consent_form)
            for key, value in self.config.user_consent_form_data.items()
        }
        ssl_verify = getattr(self.config, "ssl_verify", True)
        try:
            return self.session.post(
                self.authorization_endpoint,
                data=user_consent_data,
                headers=USER_AGENT,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
            )
        except requests.exceptions.Timeout as exc:
            raise TimeoutError(exc, "The authentication request timed out.") from exc
        except requests.RequestException as exc:
            raise RequestError.from_error(
                exc, "An error occurred while authenticating the user."
            ) from exc

    def _prepare_token_post_data(self, token_data: dict[str, Any]) -> dict[str, Any]:
        """Prepare the common data to post to the token URI"""
        token_data.update(
            {
                "redirect_uri": self.config.redirect_uri,
                "client_id": self.config.client_id,
            }
        )
        # If necessary, change the keys of the form data that will be passed to the token exchange POST request
        custom_token_exchange_params = getattr(self.config, "token_exchange_params", {})
        if custom_token_exchange_params:
            token_data[custom_token_exchange_params["redirect_uri"]] = token_data.pop(
                "redirect_uri"
            )
            token_data[custom_token_exchange_params["client_id"]] = token_data.pop(
                "client_id"
            )
        # If the client_secret is known, the token exchange request must be authenticated with a BASIC Auth, using the
        # client_id and client_secret as username and password respectively
        if getattr(self.config, "client_secret", None):
            token_data.update(
                {
                    "auth": (self.config.client_id, self.config.client_secret),
                    "client_secret": self.config.client_secret,
                }
            )
        return token_data

    def exchange_code_for_token(self, authorized_url: str, state: str) -> Response:
        """Get exchange code for token"""
        qs = parse_qs(urlparse(authorized_url).query)
        if qs["state"][0] != state:
            raise AuthenticationError(
                "The state received in the authorized url does not match initially computed state"
            )
        code = qs["code"][0]
        token_exchange_data: dict[str, Any] = {
            "code": code,
            "state": state,
            "grant_type": "authorization_code",
        }
        token_exchange_data = self._prepare_token_post_data(token_exchange_data)
        post_request_kwargs: Any = {
            self.config.token_exchange_post_data_method: token_exchange_data
        }
        ssl_verify = getattr(self.config, "ssl_verify", True)
        try:
            r = self.session.post(
                self.token_endpoint,
                headers=USER_AGENT,
                timeout=HTTP_REQ_TIMEOUT,
                verify=ssl_verify,
                **post_request_kwargs,
            )
            return r
        except requests.exceptions.Timeout as exc:
            raise TimeoutError(exc, "The authentication request timed out.") from exc
        except requests.RequestException as exc:
            raise RequestError.from_error(
                exc, "An error occurred while authenticating the user."
            ) from exc

    def _constant_or_xpath_extracted(
        self, value: str, form_element: Any
    ) -> Optional[str]:
        match = self.CONFIG_XPATH_REGEX.match(value)
        if not match:
            return value
        value_from_xpath = form_element.xpath(
            match.groupdict("xpath_value")["xpath_value"]
        )
        if len(value_from_xpath) == 1:
            return value_from_xpath[0]
        return None

    @staticmethod
    def compute_state() -> str:
        """Compute state"""
        rand = SystemRandom()
        return "".join(
            repeatfunc(
                rand.choice,
                22,
                string.digits + string.ascii_lowercase + string.ascii_uppercase,
            )
        )


class CodeAuthorizedAuth(AuthBase):
    """CodeAuthorizedAuth custom authentication class to be used with requests module"""

    def __init__(self, token: str, where: str, key: Optional[str] = None) -> None:
        self.token = token
        self.where = where
        self.key = key

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        """Perform the actual authentication"""
        if self.where == "qs":
            parts = urlparse(str(request.url))
            query_dict = parse_qs(parts.query)
            if self.key is not None:
                query_dict.update({self.key: [self.token]})
            url_without_args = parts._replace(query="").geturl()

            request.prepare_url(url_without_args, query_dict)

        elif self.where == "header":
            request.headers["Authorization"] = "Bearer {}".format(self.token)
        logger.debug(
            re.sub(
                r"'Bearer [^']+'",
                r"'Bearer ***'",
                f"PreparedRequest: {request.__dict__}",
            )
        )
        return request



# -*- coding: utf-8 -*-
# Copyright 2024, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import configparser
import os
import re
import sys
import unittest
from typing import Any, Iterator

import importlib_metadata
from packaging.requirements import Requirement
from stdlib_list import stdlib_list

from eodag.config import PluginConfig, load_default_config
from tests.context import MisconfiguredError

project_path = "./eodag"
setup_cfg_path = "./setup.cfg"
allowed_missing_imports = ["eodag"]


def get_imports(filepath: str) -> Iterator[Any]:
    """Get python imports from the given file path"""
    with open(filepath, "r") as file:
        try:
            root = ast.parse(file.read())
        except UnicodeDecodeError as e:
            raise MisconfiguredError(
                f"UnicodeDecodeError in {filepath}: {e.object[max(e.start - 50, 0):min(e.end + 50, len(e.object))]!r}"
            ) from e

    for node in ast.iter_child_nodes(root):
        if isinstance(node, ast.Import):
            for alias in node.names:
                if alias.name.split(".")[0] == "utils":
                    pass

                yield alias.name.split(".")[0]
        elif isinstance(node, ast.ImportFrom) and node.level == 0:
            if node.module.split(".")[0] == "utils":
                pass
            yield node.module.split(".")[0]


def get_project_imports(project_path: str) -> set[str]:
    """Get python imports from the project path"""
    imports: set[str] = set()
    for dirpath, dirs, files in os.walk(project_path):
        for filename in files:
            if filename.endswith(".py"):
                filepath = os.path.join(dirpath, filename)
                imports.update(get_imports(filepath))
    return imports


def get_setup_requires(setup_cfg_path: str):
    """Get requirements from the given setup.cfg file path"""
    config = configparser.ConfigParser()
    config.read(setup_cfg_path)
    return set(
        [
            Requirement(r).name
            for r in config["options"]["install_requires"].split("\n")
            if r
        ]
    )


def get_optional_dependencies(setup_cfg_path: str, extra: str) -> set[str]:
    """Get extra requirements from the given setup.cfg file path"""
    config = configparser.ConfigParser()
    config.read(setup_cfg_path)
    deps = set()
    for req in config["options.extras_require"][extra].split("\n"):
        if req.startswith("eodag["):
            for found_extra in re.findall(r"([\w-]+)[,\]]", req):
                deps.update(get_optional_dependencies(setup_cfg_path, found_extra))
        elif req:
            deps.add(Requirement(req).name)

    return deps


def get_resulting_extras(setup_cfg_path: str, extra: str) -> set[str]:
    """Get resulting extras for a single extra from the given setup.cfg file path"""
    config = configparser.ConfigParser()
    config.read(setup_cfg_path)
    extras = set()
    for req in config["options.extras_require"][extra].split("\n"):
        if req.startswith("eodag["):
            extras.update(re.findall(r"([\w-]+)[,\]]", req))
    return extras


def get_entrypoints_extras(setup_cfg_path: str) -> dict[str, str]:
    """Get entrypoints and associated extra from the given setup.cfg file path"""
    config = configparser.ConfigParser()
    config.read(setup_cfg_path)
    plugins_extras_dict = dict()
    for group in config["options.entry_points"].keys():
        for ep in config["options.entry_points"][group].split("\n"):
            # plugin entrypoint with associated extra
            match = re.search(r"^(\w+) = [\w\.:]+ \[(\w+)\]$", ep)
            if match:
                plugins_extras_dict[match.group(1)] = match.group(2)
                continue
            # plugin entrypoint without extra
            match = re.search(r"^(\w+) = [\w\.:]+$", ep)
            if match:
                plugins_extras_dict[match.group(1)] = None

    return plugins_extras_dict


class TestRequirements(unittest.TestCase):
    def test_all_requirements(self):
        """Needed libraries must be in project requirements"""

        project_imports = get_project_imports(project_path)
        setup_requires = get_setup_requires(setup_cfg_path)
        setup_requires.update(get_optional_dependencies(setup_cfg_path, "all"))
        import_required_dict = importlib_metadata.packages_distributions()
        try:
            default_libs = stdlib_list()
        except FileNotFoundError:
            # python version might not be supported by `stdlib_list`
            # Since python3.10, we can use `sys.stdlib_module_names` instead
            default_libs = list(sys.stdlib_module_names)

        missing_imports = []
        for project_import in project_imports:
            required = import_required_dict.get(project_import, [project_import])
            if (
                not set(required).intersection(setup_requires)
                and project_import not in default_libs + allowed_missing_imports
            ):
                missing_imports.append(project_import)

        self.assertEqual(
            len(missing_imports),
            0,
            f"The following libraries were not found in project requirements: {missing_imports}",
        )

    def test_plugins_extras(self):
        """All optional dependencies needed by providers must be resolved with all-providers extra"""

        plugins_extras_dict = get_entrypoints_extras(setup_cfg_path)
        all_providers_extras = get_resulting_extras(setup_cfg_path, "all-providers")

        providers_config = load_default_config()
        plugins = set()
        for provider_conf in providers_config.values():
            plugins.update(
                [
                    getattr(provider_conf, x).type
                    for x in dir(provider_conf)
                    if isinstance(getattr(provider_conf, x), PluginConfig)
                ]
            )

        for plugin in plugins:
            if extra := plugins_extras_dict.get(plugin):
                self.assertIn(extra, all_providers_extras)
